# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes that allow access to diagrams in the model."""
from __future__ import annotations

import abc
import base64
import importlib.metadata as imm
import logging
import operator
import traceback
import typing as t
import uuid

import markupsafe

import capellambse.svg

from .. import aird
from . import common as c
from . import modeltypes

DiagramConverter = t.Callable[[aird.Diagram], t.Any]
LOGGER = logging.getLogger(__name__)


class UnknownOutputFormat(ValueError):
    """An unknown output format was requested for the diagram."""


class AbstractDiagram(metaclass=abc.ABCMeta):
    """Abstract superclass of model diagrams."""

    uuid: str
    """Unique ID of this diagram."""
    name: str
    """Human-readable name for this diagram."""
    target: c.GenericElement
    """This diagram's "target".

    The target of a diagram is usually:

    *   The model element which is the direct parent of all visible
        nodes **OR**
    *   The only top-level element in the diagram **OR**
    *   The element which is considered to be the "element of interest".
    """

    _model: capellambse.MelodyModel
    _render: aird.Diagram
    _error: BaseException

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self._model = model

    def __dir__(self) -> list[str]:
        return dir(type(self)) + [
            f"as_{i.name}"
            for i in imm.entry_points()["capellambse.diagram.formats"]
        ]

    def __getattr__(self, attr: str) -> t.Any:
        if attr.startswith("as_"):
            fmt = attr[len("as_") :]
            try:
                return self.render(fmt)
            except UnknownOutputFormat:
                raise
            except Exception as err:  # pylint: disable=broad-except
                if hasattr(self, "_error") and err is self._error:
                    err_img = self._render
                else:
                    err_img = self.__create_error_image("render", err)
                assert err_img is not None
                return _find_format_converter(fmt)(err_img)
        return getattr(super(), attr)

    def __repr__(self) -> str:
        return f"<Diagram {self.name!r}>"

    def __html__(self) -> markupsafe.Markup:
        return (
            markupsafe.Markup("<figure>")
            + markupsafe.Markup(self.render("svg"))
            + markupsafe.Markup("<figcaption>")
            + self.name
            + markupsafe.Markup("</figcaption></figure>")
        )

    def _repr_svg_(self):
        return self.render("svg")

    @property
    def nodes(self) -> c.MixedElementList:
        """Return a list of all nodes visible in this diagram."""
        allids = {e.uuid for e in self.render(None)}
        assert None not in allids
        elems = []
        for elemid in allids:
            assert elemid is not None
            try:
                elem = self._model._loader[elemid]
                elem = next(elem.iterchildren("target"))
                elem = self._model._loader.follow_link(
                    elem, elem.attrib["href"]
                )
            except (KeyError, StopIteration):  # pragma: no cover
                continue
            else:
                # Filter out visual-only elements that live in the
                # .aird / .airdfragment files
                frag = self._model._loader.find_fragment(elem)
                if frag.suffix not in {".aird", ".airdfragment"}:
                    elems.append(elem)

        return c.MixedElementList(self._model, elems, c.GenericElement)

    @t.overload
    def render(self, fmt: None) -> aird.Diagram:
        ...

    @t.overload
    def render(self, fmt: str) -> t.Any:
        ...

    def render(self, fmt: str | None) -> t.Any:
        """Render the diagram in the given format."""
        conv: t.Any
        if fmt is None:
            conv = lambda i: i
        else:
            conv = _find_format_converter(fmt)

        try:
            return self.__load_cache(conv)
        except KeyError:
            pass
        return conv(self.__render_fresh())

    @abc.abstractmethod
    def _create_diagram(self) -> aird.Diagram:
        """Perform the actual rendering of the diagram.

        This method should only be called by the public :meth:`render`
        method, as it handles caching of the results.
        """

    def __create_error_image(
        self,
        stage: str,
        error: Exception,
    ) -> aird.Diagram:
        err_name = (
            "An error occured while rendering diagram\n"
            f"{self.name!r}\n"
            f"(in stage {stage!r})"
        )
        err_msg = (
            "Please report this error to your tools and methods team,"
            " and attach the following information:"
        )
        err_trace = "".join(
            traceback.format_exception(None, error, error.__traceback__)
        )

        diag = aird.Diagram("An error occured! :(")
        err_box = aird.Box(
            (200, 0),
            (350, 0),
            label=err_name,
            uuid="error",
            styleclass="Note",
            styleoverrides={
                "fill": aird.RGB(255, 0, 0),
                "text_fill": aird.RGB(255, 255, 255),
            },
        )
        diag.add_element(err_box, extend_viewport=False)
        info_box = aird.Box(
            (200, err_box.pos.y + err_box.size.y + 10),
            (350, 0),
            label=err_msg,
            uuid="info",
            styleclass="Note",
        )
        diag.add_element(info_box, extend_viewport=False)
        trace_box = aird.Box(
            (0, info_box.pos.y + info_box.size.y + 10),
            (750, 0),
            label=err_trace,
            uuid="trace",
            styleclass="Note",
        )
        diag.add_element(trace_box, extend_viewport=False)
        diag.calculate_viewport()
        return diag

    def __load_cache(self, converter):
        try:
            cache_handler = self._model._diagram_cache
            cachedir = self._model._diagram_cache_subdir
            fromcache = converter.from_cache
            ext = converter.filename_extension
        except AttributeError:
            raise KeyError(self.uuid) from None

        try:
            with cache_handler.open(cachedir / (self.uuid + ext)) as f:
                cache = f.read()
        except FileNotFoundError:
            raise KeyError(self.uuid) from None

        return fromcache(cache)

    def __render_fresh(self):
        # pylint: disable=broad-except
        if not hasattr(self, "_render"):
            try:
                self._render = self._create_diagram()
            except Exception as err:
                self._error = err
                self._render = self.__create_error_image("parse", self._error)

        if hasattr(self, "_error"):
            raise self._error
        return self._render


class Diagram(AbstractDiagram):
    """Provides access to a single diagram."""

    uuid: str = property(operator.attrgetter("_element.uid"))  # type: ignore[assignment]
    xtype = "viewpoint:DRepresentationDescriptor"
    name: str = property(operator.attrgetter("_element.name"))  # type: ignore[assignment]
    viewpoint: str = property(operator.attrgetter("_element.viewpoint"))  # type: ignore[assignment]
    target: c.GenericElement = property(lambda self: self._model.by_uuid(self.target_uuid))  # type: ignore[assignment]
    target_uuid: str = property(lambda self: self._element.target.split("#")[-1])  # type: ignore[assignment]
    """Provides slightly better performance than ``target.uuid``."""

    _element: aird.DiagramDescriptor

    @classmethod
    def from_model(
        cls,
        model: capellambse.MelodyModel,
        descriptor: aird.DiagramDescriptor,
    ) -> Diagram:
        """Wrap a diagram already defined in the Capella AIRD."""
        self = cls.__new__(cls)
        self._model = model
        self._element = descriptor
        return self

    def __init__(self, **kw: t.Any) -> None:
        # pylint: disable=super-init-not-called
        # Do I need to explain this suppression?
        raise TypeError("Cannot create a Diagram this way")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._model is other._model and self._element == other._element

    @property
    def description(self) -> str:
        """Return the diagram description."""
        desc = self._model._loader[self.uuid].get("documentation")
        return desc and c.markuptype(desc)

    @property
    def type(self) -> modeltypes.DiagramType:
        """Return the type of this diagram."""
        sc = self._element.styleclass
        try:
            return modeltypes.DiagramType(sc)
        except ValueError:  # pragma: no cover
            LOGGER.warning("Unknown diagram type %r", sc)
            return modeltypes.DiagramType.UNKNOWN

    def _create_diagram(self) -> aird.Diagram:
        return aird.parse_diagram(self._model._loader, self._element)


class DiagramAccessor(c.Accessor):
    """Provides access to a list of diagrams below the specified viewpoint."""

    def __init__(
        self,
        viewpoint: str | None = None,
        *,
        cacheattr: str | None = None,
    ) -> None:
        super().__init__()
        self.cacheattr = cacheattr
        self.viewpoint = viewpoint

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        if self.viewpoint is None:
            descriptors = list(aird.enumerate_diagrams(obj._model._loader))
        else:
            descriptors = [
                d
                for d in aird.enumerate_diagrams(obj._model._loader)
                if d.viewpoint == self.viewpoint
            ]
        return c.CachedElementList(
            obj._model,
            descriptors,
            Diagram,
            cacheattr=self.cacheattr,
        )


class SVGFormat:
    """Convert the diagram to SVG."""

    filename_extension = ".svg"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return convert_svgdiagram(diagram).to_string()

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


def convert_svgdiagram(
    diagram: aird.Diagram,
) -> capellambse.svg.generate.SVGDiagram:
    """Convert the diagram to a SVGDiagram."""
    jsondata = aird.DiagramJSONEncoder().encode(diagram)
    svg = capellambse.svg.generate.SVGDiagram.from_json(jsondata)
    return svg


class ConfluenceSVGFormat:
    """Convert the diagram to Confluence-style SVG."""

    filename_extension = ".svg"
    prefix = (
        f'<ac:structured-macro ac:macro-id="{uuid.uuid4()!s}" ac:name="html"'
        ' ac:schema-version="1">'
        f"<ac:plain-text-body><![CDATA["
    )
    postfix = "]]></ac:plain-text-body></ac:structured-macro>"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @classmethod
    def convert(cls, diagram: aird.Diagram) -> str:
        return "".join((cls.prefix, SVGFormat.convert(diagram), cls.postfix))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        return "".join((cls.prefix, cache.decode("utf-8"), cls.postfix))


class SVGDataURIFormat:
    filename_extension = ".svg"
    preamble = "data:image/svg+xml;base64,"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @classmethod
    def convert(cls, diagram: aird.Diagram) -> str:
        """Convert the diagram to a ``data:`` URI containing SVG."""
        payload = SVGFormat.convert(diagram)
        b64 = base64.standard_b64encode(payload.encode("utf-8"))
        return "".join((cls.preamble, b64.decode("ascii")))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        b64 = base64.standard_b64encode(cache)
        return "".join((cls.preamble, b64.decode("ascii")))


class SVGInHTMLIMGFormat:
    filename_extension = ".svg"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        """Convert the diagram to an HTML ``<img>``."""
        payload = SVGDataURIFormat.convert(diagram)
        return c.markuptype(f'<img src="{payload}"/>')

    @staticmethod
    def from_cache(cache: bytes) -> str:
        payload = SVGDataURIFormat.from_cache(cache)
        return c.markuptype(f'<img src="{payload}"/>')


class JSONFormat:
    filename_extension = ".json"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return aird.DiagramJSONEncoder().encode(diagram)

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


class PrettyJSONFormat:
    filename_extension = ".json"

    def __new__(cls, diagram):
        return cls.convert(diagram)

    @staticmethod
    def convert(diagram: aird.Diagram) -> str:
        return aird.DiagramJSONEncoder(indent=4).encode(diagram)

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


def _find_format_converter(fmt: str) -> DiagramConverter:
    try:
        return next(
            i.load()
            for i in imm.entry_points()["capellambse.diagram.formats"]
            if i.name == fmt
        )
    except StopIteration:
        raise UnknownOutputFormat(
            f"Unknown image output format {fmt}"
        ) from None



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for handling ReqIF Requirements.

.. diagram:: [CDB] Requirements ORM
"""
from __future__ import annotations

import collections.abc as cabc
import datetime
import logging
import typing as t

from lxml import etree

import capellambse.model
import capellambse.model.common as c
from capellambse import helpers
from capellambse.loader import xmltools
from capellambse.model import crosslayer

XT_REQUIREMENT = "Requirements:Requirement"
XT_REQ_ATTR_STRINGVALUE = "Requirements:StringValueAttribute"
XT_REQ_ATTR_REALVALUE = "Requirements:RealValueAttribute"
XT_REQ_ATTR_INTEGERVALUE = "Requirements:IntegerValueAttribute"
XT_REQ_ATTR_DATEVALUE = "Requirements:DateValueAttribute"
XT_REQ_ATTR_BOOLEANVALUE = "Requirements:BooleanValueAttribute"
XT_REQ_ATTR_ENUMVALUE = "Requirements:EnumerationValueAttribute"
XT_REQ_ATTRIBUTES = {
    XT_REQ_ATTR_ENUMVALUE,
    XT_REQ_ATTR_STRINGVALUE,
    XT_REQ_ATTR_REALVALUE,
    XT_REQ_ATTR_INTEGERVALUE,
    XT_REQ_ATTR_DATEVALUE,
    XT_REQ_ATTR_BOOLEANVALUE,
}
XT_INC_RELATION = "CapellaRequirements:CapellaIncomingRelation"
XT_OUT_RELATION = "CapellaRequirements:CapellaOutgoingRelation"
XT_INT_RELATION = "Requirements:InternalRelation"
XT_MODULE = "CapellaRequirements:CapellaModule"
XT_FOLDER = "Requirements:Folder"

XT_REQ_TYPES_F = "CapellaRequirements:CapellaTypesFolder"
XT_REQ_TYPES_DATA_DEF = "Requirements:DataTypeDefinition"
XT_REQ_TYPE = "Requirements:RequirementType"
XT_RELATION_TYPE = "Requirements:RelationType"
XT_MODULE_TYPE = "Requirements:ModuleType"
XT_REQ_TYPE_ENUM = "Requirements:EnumerationDataTypeDefinition"
XT_REQ_TYPE_ATTR_ENUM = "Requirements:EnumValue"
XT_REQ_TYPE_ATTR_DEF = "Requirements:AttributeDefinition"
XT_REQ_TYPE_ENUM_DEF = "Requirements:AttributeDefinitionEnumeration"
XT_REQ_TYPES = {
    XT_REQ_TYPES_F,
    XT_REQ_TYPES_DATA_DEF,
    XT_REQ_TYPE,
    XT_RELATION_TYPE,
    XT_MODULE_TYPE,
    XT_REQ_TYPE_ENUM,
    XT_REQ_TYPE_ATTR_ENUM,
    XT_REQ_TYPE_ATTR_DEF,
    XT_REQ_TYPE_ENUM_DEF,
}
DATE_VALUE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

logger = logging.getLogger("reqif")
_xt_to_attr_type: dict[
    str, tuple[cabc.Callable[[str], t.Any], type | tuple[type, ...], t.Any]
] = {
    XT_REQ_ATTR_STRINGVALUE: (str, str, ""),
    XT_REQ_ATTR_REALVALUE: (float, float, 0.0),
    XT_REQ_ATTR_INTEGERVALUE: (int, int, 0),
    XT_REQ_ATTR_DATEVALUE: (
        lambda val: datetime.datetime.strptime(val, DATE_VALUE_FORMAT),
        (datetime.datetime, type(None)),
        None,
    ),
    XT_REQ_ATTR_BOOLEANVALUE: (lambda val: val == "true", bool, False),
}
_attr_type_hints = {
    "int": XT_REQ_ATTR_INTEGERVALUE,
    "integer": XT_REQ_ATTR_INTEGERVALUE,
    "integervalueattribute": XT_REQ_ATTR_INTEGERVALUE,
    "str": XT_REQ_ATTR_STRINGVALUE,
    "string": XT_REQ_ATTR_STRINGVALUE,
    "stringvalueattribute": XT_REQ_ATTR_STRINGVALUE,
    "float": XT_REQ_ATTR_REALVALUE,
    "real": XT_REQ_ATTR_REALVALUE,
    "realvalueattribute": XT_REQ_ATTR_REALVALUE,
    "date": XT_REQ_ATTR_DATEVALUE,
    "datevalueattribute": XT_REQ_ATTR_DATEVALUE,
    "bool": XT_REQ_ATTR_BOOLEANVALUE,
    "boolvalueattribute": XT_REQ_ATTR_BOOLEANVALUE,
    "enum": XT_REQ_ATTR_ENUMVALUE,
    "enumvalueattribute": XT_REQ_ATTR_ENUMVALUE,
}


class RequirementsRelationAccessor(
    c.WritableAccessor["AbstractRequirementsRelation"]
):
    """Searches for requirement relations in the architecture layer."""

    # pylint: disable=abstract-method  # Only partially implemented for now

    __slots__ = ("aslist",)

    def __init__(self, *args, **kw) -> None:
        super().__init__(*args, **kw, aslist=c.ElementList)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        rel_objs = list(
            obj._model._loader.iterchildren_xt(obj._element, XT_INC_RELATION)
        )

        for i in obj._model._loader.iterall_xt(XT_OUT_RELATION):
            if RequirementsOutRelation.from_model(obj._model, i).source == obj:
                rel_objs.append(i)

        for i in obj._model._loader.iterall_xt(XT_INT_RELATION):
            rel = RequirementsIntRelation.from_model(obj._model, i)
            if obj in (rel.source, rel.target):
                rel_objs.append(i)
        return self._make_list(obj, rel_objs)

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model, elements, c.GenericElement, parent=parent_obj
        )

    def create(
        self,
        elmlist: c.ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> RequirementsIntRelation | RequirementsIncRelation:
        if "target" not in kw:
            raise TypeError("No `target` for new requirement relation")
        cls, xtype = self._find_relation_type(kw["target"])
        parent = elmlist._parent._element
        with elmlist._model._loader.new_uuid(parent) as uuid:
            return cls(
                elmlist._model,
                parent,
                **kw,
                source=elmlist._parent,
                uuid=uuid,
                xtype=xtype,
            )

    def _find_relation_type(
        self, target: c.GenericElement
    ) -> tuple[type[RequirementsIntRelation | RequirementsIncRelation], str]:
        if isinstance(target, Requirement):
            return (
                RequirementsIntRelation,
                XT_INT_RELATION,
            )
        elif isinstance(target, ReqIFElement):
            raise TypeError(
                "Cannot create relations to targets of type"
                f" {type(target).__name__}"
            )
        else:
            return (
                RequirementsIncRelation,
                XT_INC_RELATION,
            )


class ElementRelationAccessor(
    c.WritableAccessor["AbstractRequirementsRelation"]
):
    """Provides access to RequirementsRelations of a GenericElement."""

    # pylint: disable=abstract-method  # Only partially implemented for now

    __slots__ = ("aslist",)

    def __init__(self) -> None:
        super().__init__(aslist=RelationsList)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        relations: list[etree._Element] = []
        for relation in obj._model.search(
            XT_INC_RELATION, XT_INT_RELATION, XT_OUT_RELATION
        ):
            if obj in (relation.source, relation.target):
                relations.append(relation._element)
        return self._make_list(obj, relations)

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model,
            elements,
            None,
            parent=parent_obj,
            source=parent_obj,
        )


class ReqIFElement(c.GenericElement):
    """Attributes shared by all ReqIF elements."""

    identifier = xmltools.AttributeProperty(
        "_element", "ReqIFIdentifier", optional=True
    )
    long_name = xmltools.AttributeProperty(
        "_element", "ReqIFLongName", optional=True
    )
    description = xmltools.AttributeProperty(
        "_element", "ReqIFDescription", optional=True
    )
    name = xmltools.AttributeProperty("_element", "ReqIFName", optional=True)
    prefix = xmltools.AttributeProperty(
        "_element", "ReqIFPrefix", optional=True
    )
    type: c.Accessor = property(lambda _: None)  # type: ignore[assignment]

    def __repr__(self) -> str:  # pragma: no cover
        mytype = type(self).__name__
        path = []
        parent = self._element
        if isinstance(
            self,
            (
                RequirementsOutRelation,
                RequirementsIncRelation,
                RequirementsIntRelation,
            ),
        ):
            return (
                f"<{mytype} from {self.source!r} to {self.target!r} "
                f"({self.uuid})>"
            )
        elif self.xtype in XT_REQ_TYPES:
            return f'<{mytype} {parent.get("ReqIFLongName")!r} ({self.uuid})>'
        while parent is not None:
            path.append(
                parent.get("ReqIFText")
                or parent.get("ReqIFName")
                or parent.get("ReqIFChapterName")
                or parent.get("ReqIFLongName")
                or "..."
            )
            if helpers.xtype_of(parent) == XT_MODULE:
                break
            parent = parent.getparent()

        return f'<{mytype} {"/".join(reversed(path))!r} ({self.uuid})>'


@c.xtype_handler(None, XT_REQ_TYPES_DATA_DEF)
class DataTypeDefinition(ReqIFElement):
    """A data type definition for requirement types"""

    _xmltag = "ownedDefinitionTypes"


@c.xtype_handler(None, XT_REQ_TYPE_ATTR_DEF)
class AttributeDefinition(ReqIFElement):
    """An attribute definition for requirement types"""

    _xmltag = "ownedAttributes"

    data_type = c.AttrProxyAccessor(DataTypeDefinition, "definitionType")


class AbstractRequirementsAttribute(c.GenericElement):
    _xmltag = "ownedAttributes"

    definition: t.ClassVar[c.Accessor]

    def __repr__(self) -> str:
        mytype = self.xtype.rsplit(":", maxsplit=1)[-1]
        try:
            name = self.definition.long_name
        except AttributeError:
            name = ""
        return f"<{mytype} [{name}] ({self.uuid})>"


class AttributeAccessor(
    c.ProxyAccessor["RequirementsAttribute | EnumerationValueAttribute"]
):
    def __init__(self) -> None:
        super().__init__(
            c.GenericElement,  # type: ignore[arg-type]
            XT_REQ_ATTRIBUTES,
            aslist=c.MixedElementList,
            list_extra_args={
                "mapkey": "definition.long_name",
                "mapvalue": "value",
            },
        )

    def _match_xtype(self, type_: str) -> tuple[type, str]:  # type: ignore[override]
        type_ = type_.lower()
        try:
            xt = _attr_type_hints[type_]
        except KeyError:
            raise ValueError(f"Invalid type hint given: {type_!r}") from None

        if xt == XT_REQ_ATTR_ENUMVALUE:
            return EnumerationValueAttribute, xt

        return RequirementsAttribute, xt


class RelationsList(c.ElementList["AbstractRequirementsRelation"]):
    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[t.Any] | None = None,
        *,
        source: c.ModelObject,
    ) -> None:
        del elemclass
        super().__init__(model, elements, c.GenericElement)  # type: ignore[arg-type]
        self._source = source

    @t.overload
    def __getitem__(self, idx: int | str) -> c.T:
        ...

    @t.overload
    def __getitem__(self, idx: slice) -> c.ElementList[c.T]:
        ...

    def __getitem__(self, idx):
        rel = super().__getitem__(idx)
        if isinstance(rel, c.ElementList):
            return rel

        assert isinstance(rel, AbstractRequirementsRelation)
        assert self._source in (rel.source, rel.target)
        if self._source == rel.source:
            return rel.target
        else:
            return rel.source

    def by_relation_type(self, reltype: str) -> RelationsList:
        matches = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            assert isinstance(rel_elm, AbstractRequirementsRelation)
            if rel_elm.type is not None and rel_elm.type.name == reltype:
                matches.append(elm)
        return self._newlist(matches)

    def by_relation_class(
        self, class_: t.Literal["incoming", "outgoing", "internal"]
    ) -> RelationsList:
        relation_types = {
            "incoming": RequirementsIncRelation,
            "outgoing": RequirementsOutRelation,
            "internal": RequirementsIntRelation,
        }
        matches: list[etree._Element] = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            if isinstance(rel_elm, relation_types[class_]):
                matches.append(rel_elm._element)
        return self._newlist(matches)

    def _newlist(self, elements: list[etree._Element]) -> RelationsList:
        listtype = self._newlist_type()
        assert issubclass(listtype, RelationsList)
        return listtype(self._model, elements, source=self._source)


class ValueAccessor(c.Accessor):
    def __get__(  # type: ignore[override]
        self,
        obj: c.GenericElement | None,
        objtype: type[ReqIFElement] | None = None,
    ) -> ValueAccessor | int | float | str | bool | datetime.datetime:
        del objtype
        if obj is None:
            return self

        assert isinstance(
            obj, (RequirementsAttribute, EnumerationValueAttribute)
        )
        cast, _, default = _xt_to_attr_type[obj.xtype]
        if (value := obj._element.get("value")) is None:
            return default
        return cast(value)

    def __set__(
        self,
        obj: c.GenericElement,
        value: int | float | str | bool | datetime.datetime,
    ) -> None:
        assert isinstance(
            obj, (RequirementsAttribute, EnumerationValueAttribute)
        )
        _, type_, default = _xt_to_attr_type[obj.xtype]
        if not isinstance(value, type_):
            if isinstance(type_, tuple):
                error_msg = " or ".join((t.__name__ for t in type_))
            else:
                error_msg = type_.__name__

            raise TypeError("Value needs to be of type " + error_msg)

        if value == default:
            if "value" in obj._element.attrib:
                del obj._element.attrib["value"]
            return

        if isinstance(value, bool):
            value = str(value).lower()
        elif isinstance(value, datetime.datetime):
            value = datetime.datetime.strftime(value, DATE_VALUE_FORMAT)
        else:
            value = str(value)

        obj._element.attrib["value"] = value


@c.xtype_handler(None, *(XT_REQ_ATTRIBUTES - {XT_REQ_ATTR_ENUMVALUE}))
class RequirementsAttribute(AbstractRequirementsAttribute):
    """Handles attributes on Requirements."""

    definition = c.AttrProxyAccessor(AttributeDefinition, "definition")

    value = ValueAccessor()


@c.xtype_handler(None, XT_REQ_TYPE_ATTR_ENUM)
class EnumValue(ReqIFElement):
    """An enumeration value for :class:`EnumDataTypeDefinition`"""

    _xmltag = "specifiedValues"

    def __str__(self) -> str:
        return self.long_name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.long_name == other
        return super().__eq__(other)


@c.xtype_handler(None, XT_REQ_TYPE_ENUM)
class EnumDataTypeDefinition(ReqIFElement):
    """An enumeration data type definition for requirement types"""

    _xmltag = "ownedDefinitionTypes"

    values = c.ProxyAccessor(
        EnumValue, XT_REQ_TYPE_ATTR_ENUM, aslist=c.ElementList
    )


@c.xtype_handler(None, XT_REQ_TYPE_ENUM_DEF)
class AttributeDefinitionEnumeration(ReqIFElement):
    """An enumeration attribute definition for requirement types"""

    _xmltag = "enumeration"

    data_type = c.AttrProxyAccessor(EnumDataTypeDefinition, "definitionType")
    multi_valued = xmltools.BooleanAttributeProperty(
        "_element",
        "multiValued",
        __doc__="Boolean flag for setting multiple enumeration values on the attribute",
    )


@c.xtype_handler(None, XT_REQ_ATTR_ENUMVALUE)
class EnumerationValueAttribute(AbstractRequirementsAttribute):
    """An enumeration attribute."""

    definition = c.AttrProxyAccessor(
        AttributeDefinitionEnumeration, "definition"
    )
    values = c.AttrProxyAccessor(EnumValue, "values", aslist=c.ElementList)

    @property
    def value(self):
        vals = self.values
        if len(vals) > 1:
            raise TypeError("Multi-value enumeration, use `.values` instead")
        if len(vals) == 1:
            return vals[0]
        return None  # TODO return a default value?


class AbstractType(ReqIFElement):
    owner = c.ParentAccessor(c.GenericElement)
    attribute_definitions = c.ProxyAccessor(
        c.GenericElement,
        (XT_REQ_TYPE_ATTR_DEF, XT_REQ_TYPE_ENUM_DEF),
        aslist=c.MixedElementList,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.long_name == other

        return super().__eq__(other)


@c.xtype_handler(None, XT_MODULE_TYPE)
class ModuleType(AbstractType):
    """A requirement-module type"""

    _xmltag = "ownedTypes"


@c.xtype_handler(None, XT_RELATION_TYPE)
class RelationType(AbstractType):
    """A requirement-relation type"""

    _xmltag = "ownedTypes"


@c.xtype_handler(None, XT_REQ_TYPE)
class RequirementType(AbstractType):
    """A requirement type"""

    _xmltag = "ownedTypes"

    long_name = xmltools.AttributeProperty(
        "_element", "ReqIFLongName", optional=True
    )


@c.xtype_handler(None, XT_REQUIREMENT)
class Requirement(ReqIFElement):
    """A ReqIF Requirement."""

    _xmltag = "ownedRequirements"

    owner = c.ParentAccessor(c.GenericElement)

    chapter_name = xmltools.AttributeProperty(
        "_element", "ReqIFChapterName", optional=True
    )
    foreign_id = xmltools.AttributeProperty(
        "_element", "ReqIFForeignID", optional=True, returntype=int
    )
    text = xmltools.AttributeProperty(
        "_element", "ReqIFText", optional=True, returntype=c.markuptype
    )
    attributes = AttributeAccessor()
    relations = RequirementsRelationAccessor()
    related = ElementRelationAccessor()
    type = c.AttrProxyAccessor(RequirementType, "requirementType")


@c.xtype_handler(None, XT_FOLDER)
class RequirementsFolder(Requirement):
    """A folder that stores Requirements."""

    _xmltag = "ownedRequirements"

    folders: c.Accessor
    requirements = c.ProxyAccessor(
        Requirement, XT_REQUIREMENT, aslist=c.ElementList
    )


@c.xtype_handler(None, XT_MODULE)
class RequirementsModule(ReqIFElement):
    """A ReqIF Module that bundles multiple Requirement folders."""

    _xmltag = "ownedExtensions"

    folders = c.ProxyAccessor(
        RequirementsFolder, XT_FOLDER, aslist=c.ElementList
    )
    requirements = c.ProxyAccessor(
        Requirement, XT_REQUIREMENT, aslist=c.ElementList
    )
    type = c.AttrProxyAccessor(ModuleType, "moduleType")
    attributes = AttributeAccessor()


class AbstractRequirementsRelation(ReqIFElement):
    _required_attrs = frozenset({"source", "target"})

    type = c.AttrProxyAccessor(RelationType, "relationType")
    source = c.AttrProxyAccessor(Requirement, "source")
    target = c.AttrProxyAccessor(c.GenericElement, "target")


@c.xtype_handler(None, XT_OUT_RELATION)
class RequirementsOutRelation(AbstractRequirementsRelation):
    """A Relation between an object and a requirement."""

    _xmltag = "ownedExtensions"

    source = c.AttrProxyAccessor(Requirement, "target")
    target = c.AttrProxyAccessor(c.GenericElement, "source")


@c.xtype_handler(None, XT_INC_RELATION)
class RequirementsIncRelation(AbstractRequirementsRelation):
    """A Relation between a requirement and an object."""

    _xmltag = "ownedRelations"


@c.xtype_handler(None, XT_INT_RELATION)
class RequirementsIntRelation(AbstractRequirementsRelation):
    """A Relation between two requirements."""

    _xmltag = "ownedRelations"


@c.xtype_handler(None, XT_REQ_TYPES_F)
class RequirementsTypesFolder(ReqIFElement):
    _xmltag = "ownedExtensions"

    data_type_definitions = c.ProxyAccessor(
        c.GenericElement,
        (XT_REQ_TYPES_DATA_DEF, XT_REQ_TYPE_ENUM),
        aslist=c.MixedElementList,
    )
    module_types = c.ProxyAccessor(
        ModuleType, XT_MODULE_TYPE, aslist=c.ElementList
    )
    relation_types = c.ProxyAccessor(
        RelationType, XT_RELATION_TYPE, aslist=c.ElementList
    )
    requirement_types = c.ProxyAccessor(
        RequirementType, XT_REQ_TYPE, aslist=c.ElementList
    )


def init() -> None:
    c.set_accessor(
        RequirementsFolder,
        "folders",
        c.ProxyAccessor(RequirementsFolder, XT_FOLDER, aslist=c.ElementList),
    )
    c.set_accessor(c.GenericElement, "requirements", ElementRelationAccessor())
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_modules",
        c.ProxyAccessor(RequirementsModule, XT_MODULE, aslist=c.ElementList),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirements",
        c.ProxyAccessor(
            Requirement,
            XT_REQUIREMENT,
            aslist=c.ElementList,
            rootelem=XT_MODULE,
            deep=True,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_types_folders",
        c.ProxyAccessor(
            RequirementsTypesFolder,
            XT_REQ_TYPES_F,
            aslist=c.ElementList,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirement_types",
        c.ProxyAccessor(
            RequirementType,
            XT_REQ_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
            deep=True,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_module_types",
        c.ProxyAccessor(
            ModuleType,
            XT_MODULE_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
            deep=True,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_relation_types",
        c.ProxyAccessor(
            RelationType,
            XT_RELATION_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
            deep=True,
        ),
    )



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import collections.abc as cabc

from capellambse.loader import xmltools

from .. import common as c
from .. import modeltypes
from . import information

XT_COMP_EX_FNC_EX_ALLOC = "org.polarsys.capella.core.data.fa:ComponentExchangeFunctionalExchangeAllocation"
XT_FCALLOC = "org.polarsys.capella.core.data.fa:ComponentFunctionalAllocation"
XT_FCI: cabc.Set[str] = frozenset(
    {
        "org.polarsys.capella.core.data.fa:FunctionalChainInvolvementFunction",
        "org.polarsys.capella.core.data.fa:FunctionalChainInvolvementLink",
    }
)


@c.xtype_handler(None)
class FunctionRealization(c.GenericElement):
    """A realization that links to a function."""

    _xmltag = "ownedFunctionRealizations"


class AbstractExchange(c.GenericElement):
    """Common code for Exchanges."""

    source_port = c.AttrProxyAccessor(c.GenericElement, "source")
    target_port = c.AttrProxyAccessor(c.GenericElement, "target")


@c.xtype_handler(None)
class FunctionInputPort(c.GenericElement):
    """A function input port."""

    _xmltag = "inputs"

    owner = c.ParentAccessor(c.GenericElement)

    exchanges: c.Accessor


@c.xtype_handler(None)
class FunctionOutputPort(c.GenericElement):
    """A function output port."""

    _xmltag = "outputs"

    owner = c.ParentAccessor(c.GenericElement)

    exchanges: c.Accessor


@c.xtype_handler(None)
class FunctionalExchange(AbstractExchange):
    """A functional exchange."""

    _xmltag = "ownedFunctionalExchanges"

    exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem, "exchangedItems", aslist=c.ElementList
    )


@c.xtype_handler(None)
class FunctionalChain(c.GenericElement):
    """A functional chain."""


@c.xtype_handler(None)
class ComponentPort(c.GenericElement):
    """A component port."""

    _xmltag = "ownedFeatures"

    direction = xmltools.EnumAttributeProperty(
        "_element", "orientation", modeltypes.FPortDir, writable=False
    )
    owner = c.ParentAccessor(c.GenericElement)
    exchanges: c.Accessor


@c.xtype_handler(None)
class ComponentExchange(AbstractExchange):
    """A functional component exchange."""

    _xmltag = "ownedComponentExchanges"

    func_exchanges = c.ProxyAccessor(
        FunctionalExchange,
        XT_COMP_EX_FNC_EX_ALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )
    allocated_exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem,
        "convoyedInformations",
        aslist=c.ElementList,
    )

    @property
    def exchange_items(
        self,
    ) -> c.ElementList[information.ExchangeItem]:
        items = self.allocated_exchange_items
        func_exchanges = self.func_exchanges
        assert isinstance(func_exchanges, cabc.Iterable)
        for exchange in func_exchanges:
            items += exchange.exchange_items
        return items


for _port, _exchange in [
    (ComponentPort, ComponentExchange),
    (FunctionInputPort, FunctionalExchange),
    (FunctionOutputPort, FunctionalExchange),
]:
    c.set_accessor(
        _port,
        "exchanges",
        c.ReferenceSearchingAccessor(
            _exchange,
            "source_port",
            "target_port",
            aslist=c.ElementList,
        ),
    )
del _port, _exchange



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions implementing various filters that Capella supports."""
from __future__ import annotations

import collections.abc as cabc
import importlib
import pkgutil
import typing as t
import urllib.parse

from lxml import etree

import capellambse.loader
from capellambse import aird

from .. import _common as c

Phase2CompositeFilter = t.Callable[
    [c.ElementBuilder, aird.DiagramElement], None
]
CompositeFilter = t.Callable[
    [c.ElementBuilder, aird.DiagramElement], t.Optional[Phase2CompositeFilter]
]
#: Maps composite filter names to phase-1 callables
COMPOSITE_FILTERS: dict[str, CompositeFilter] = {}

GlobalFilter = t.Callable[
    [
        aird.Diagram,
        etree._Element,
        etree._Element,
        capellambse.loader.MelodyLoader,
    ],
    None,
]
#: Maps names of global filters to functions that implement them
GLOBAL_FILTERS: dict[str, GlobalFilter] = {}

_TDiagramElement = t.TypeVar("_TDiagramElement", bound=aird.DiagramElement)


def composite_filter(
    name: str,
) -> cabc.Callable[[CompositeFilter], CompositeFilter]:
    """Register a composite filter.

    Composite filters are executed in two phases.  The first phase
    occurs during each diagram element's creation.  Here the callables
    registered with this decorator are directly executed with the
    diagram under construction and the to-be-filtered element as
    arguments.

    The second phase happens after the entire diagram has been created.
    Filters that want to run in this phase should return another
    callable object during the first phase; this object will then be
    called with the same (now constructed) diagram and the element.

    Composite filters should always operate in place.  They may append
    additional elements to the diagram, but they should never remove
    any; instead they should simply change their "hidden" flag to True.
    """

    def add_comp_filter(func: CompositeFilter) -> CompositeFilter:
        COMPOSITE_FILTERS[name] = func
        return func

    return add_comp_filter


def phase2_composite_filter(
    name: str,
) -> cabc.Callable[[Phase2CompositeFilter], Phase2CompositeFilter]:
    """Register a composite filter that only needs to run in phase 2."""

    def add_comp_filter(func: Phase2CompositeFilter) -> Phase2CompositeFilter:
        def phase1dummy(
            _1: c.ElementBuilder,
            _2: aird.DiagramElement,
        ) -> Phase2CompositeFilter:
            return func

        COMPOSITE_FILTERS[name] = phase1dummy
        return func

    return add_comp_filter


def global_filter(name: str) -> cabc.Callable[[GlobalFilter], GlobalFilter]:
    """Register a global filter."""

    def add_global_filter(func: GlobalFilter) -> GlobalFilter:
        GLOBAL_FILTERS[name] = func
        return func

    return add_global_filter


def setfilters(
    seb: c.SemanticElementBuilder,
    dgobject: _TDiagramElement,
) -> _TDiagramElement:
    """Set the filters on the element.

    This is phase 1 of composite filter execution.

    Parameters
    ----------
    seb
        The element builder instance
    dgobject
        The constructed diagram element on which to set filters.

    Returns
    -------
    dgobject
        The modified ``dgobject``.  This is done for convenience; the
        object is modified in place.

    See Also
    --------
    composite_filter : Registering a composite filter.
    """
    dgobject.hidden = seb.diag_element.attrib.get("visible", "true") != "true"
    dgobject.hidelabel = False

    for flt in seb.diag_element.iterchildren("graphicalFilters"):
        flttype = flt.attrib[c.ATT_XMT]
        if flttype == "diagram:HideLabelFilter":
            dgobject.hidelabel = True
        elif flttype in {"diagram:HideFilter", "diagram:CollapseFilter"}:
            dgobject.hidden = True
        elif flttype == "diagram:AppliedCompositeFilters":
            _set_composite_filter(seb, dgobject, flt)
        else:
            c.LOGGER.warning("Ignoring unknown filter type %r", flttype)
    return dgobject


def applyfilters(
    target_diagram: aird.Diagram,
    diagram_root: etree._Element,
    melodyloader: capellambse.loader.MelodyLoader,
) -> None:
    """Apply filters on the ``target_diagram``.

    This function performs two tasks.

    Firstly it executes phase 2 of all elements' composite filters; see
    :func:`composite_filter` for more details.

    Secondly it applies the diagram's global filters.  These are always
    applied after all composite filters have run.

    Parameters
    ----------
    target_diagram
        The diagram.
    diagram_root
        The LXML element that is this diagram's root.
    melodyloader
        The MelodyLoader instance.
    """
    # Apply post-processing filters on elements
    for dgobject in target_diagram:
        try:
            filters = dgobject._compfilters  # type: ignore[union-attr]
        except AttributeError:
            continue
        assert dgobject.uuid is not None

        flttype: str
        p2flt: Phase2CompositeFilter
        for flttype, p2flt in filters:
            c.LOGGER.debug(
                "Applying post-processing filter %r to %s %r",
                flttype,
                dgobject.styleclass or dgobject.__class__.__name__,
                dgobject.uuid,
            )
            data_element = melodyloader[dgobject.uuid]
            p2flt(
                c.ElementBuilder(
                    target_diagram=target_diagram,
                    diagram_tree=diagram_root,
                    data_element=data_element,
                    melodyloader=melodyloader,
                    fragment=melodyloader.find_fragment(data_element),
                ),
                dgobject,
            )

        del dgobject._compfilters  # type: ignore[union-attr]

    # Apply global diagram filters
    for flt in diagram_root.iterchildren("activatedFilters"):
        try:
            flttype = flt.attrib[c.ATT_XMT]
        except KeyError:
            flttype = "(no filter type given)"
        if flttype != "filter:CompositeFilterDescription":
            c.LOGGER.warning("Unknown global filter type %r", flttype)
            continue

        try:
            fltname = _extract_filter_type(flt)
        except ValueError as err:
            c.LOGGER.warning("Ignoring broken global filter: %s", err)
            continue

        try:
            fltfunc = GLOBAL_FILTERS[fltname]
        except KeyError:
            c.LOGGER.warning("Unknown global filter %r", fltname)
            continue

        c.LOGGER.debug("Applying global filter %r", fltname)
        fltfunc(target_diagram, diagram_root, flt, melodyloader)


def _set_composite_filter(
    seb: c.SemanticElementBuilder,
    dgobject: aird.DiagramElement,
    flt: etree._Element,
) -> None:
    try:
        flttype = _extract_filter_type(
            next(flt.iterchildren("compositeFilterDescriptions"))
        )
    except (StopIteration, ValueError):
        c.LOGGER.warning(
            "Ignoring unusable composite filter on object %r", dgobject.uuid
        )
        return

    try:
        fltfunc = COMPOSITE_FILTERS[flttype]
    except KeyError:
        c.LOGGER.warning(
            "Unknown composite filter %r on object %r", flttype, dgobject.uuid
        )
        return

    phase2 = fltfunc(seb, dgobject)

    # Register the phase 2 callable
    if phase2 is None:
        return
    if not hasattr(dgobject, "_compfilters"):
        dgobject._compfilters = []  # type: ignore[union-attr]
    dgobject._compfilters.append((flttype, phase2))  # type: ignore[union-attr]


def _extract_filter_type(flt_elm: etree._Element) -> str:
    try:
        flttype = flt_elm.attrib["href"]
    except KeyError:
        raise ValueError("Filter element has no href") from None

    flttype = c.RE_COMPOSITE_FILTER.search(flttype)
    if not flttype or not flttype.group(1):
        raise ValueError("Filter href does not match known pattern") from None

    return urllib.parse.unquote(flttype.group(1))


# Load filter modules
for module in ("composite", "global"):
    try:
        importlib.import_module(f"{__name__}.{module}")
    except Exception as _err:  # pylint: disable=broad-except
        c.LOGGER.error(
            "Cannot load filters from %s: %s: %s",
            module,
            _err.__class__.__name__,
            _err,
        )
del module



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Python capellambse package."""
import platformdirs


def migrate_cache_dir():
    """Migrate the cache directory from the old name to the new, shorter one.

    After some period of time to allow all active installations to
    migrate, this functionality should be removed again.

    Note that this function must be the first thing that happens in the
    top-level ``__init__.py``, so that submodules and external modules
    that use it will use the new name with migrated data.
    """
    # pylint: disable=import-outside-toplevel  # Reduce namespace pollution
    import pathlib
    import sys

    olddirs = platformdirs.PlatformDirs("python-capella-mbse")
    oldcachedir = pathlib.Path(olddirs.user_cache_dir)
    newcachedir = pathlib.Path(dirs.user_cache_dir)
    if not newcachedir.exists() and oldcachedir.exists():
        try:
            oldcachedir.rename(newcachedir)
        except OSError as err:
            print(
                f"Warning: Cannot migrate cache directory to {newcachedir}:",
                f"Exception occurred: {type(err).__name__}: {err}",
                f"Please delete the old directory manually: {oldcachedir}",
                file=sys.stderr,
                sep="\n",
            )


dirs = platformdirs.PlatformDirs("capellambse")
migrate_cache_dir()
del platformdirs, migrate_cache_dir

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


from ._namespaces import (
    NAMESPACES,
    check_plugin_version,
    yield_key_and_version_from_namespaces_by_plugin,
)
from .model import MelodyModel


def load_model_extensions() -> None:
    """Load all model extensions.

    This function loads all entry points in the group
    ``capellambse.model_extensions`` and executes them.

    Note that this function must be placed at the end of the top-level
    ``__init__.py``, in order to ensure that all submodules were
    initialized before loading any extensions.
    """
    # pylint: disable=import-outside-toplevel  # Reduce namespace pollution
    import importlib.metadata as imm
    import logging

    try:
        entrypoints = imm.entry_points()["capellambse.model_extensions"]
    except KeyError:
        return

    for entrypoint in entrypoints:
        try:
            initfunc = entrypoint.load()
            initfunc()
        except Exception:
            logging.getLogger(__name__).exception(
                "Cannot load model extension %r from %r",
                entrypoint,
                entrypoint.value,
            )


load_model_extensions()
del load_model_extensions



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for the Physical Architecture layer.

.. diagram:: [CDB] PA ORM
"""
from __future__ import annotations

import operator

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, cs, fa
from . import la

XT_ARCH = "org.polarsys.capella.core.data.pa:PhysicalArchitecture"

XT_LA_COMP_REAL = (
    "org.polarsys.capella.core.data.pa:LogicalComponentRealization"
)


@c.xtype_handler(XT_ARCH)
class PhysicalFunction(c.GenericElement):
    """A physical function on the Physical Architecture layer."""

    _xmltag = "ownedPhysicalFunctions"

    inputs = c.ProxyAccessor(fa.FunctionInputPort, aslist=c.ElementList)
    outputs = c.ProxyAccessor(fa.FunctionOutputPort, aslist=c.ElementList)
    owner = c.CustomAccessor(
        c.GenericElement,
        operator.attrgetter("_model.pa.all_components"),
        matchtransform=operator.attrgetter("functions"),
    )
    realized_logical_functions = c.ProxyAccessor(
        la.LogicalFunction,
        fa.FunctionRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )

    is_leaf = property(lambda self: not self.functions)

    functions: c.Accessor


@c.xtype_handler(XT_ARCH)
class PhysicalFunctionPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedFunctionPkg"

    components = c.ProxyAccessor(PhysicalFunction, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class PhysicalComponent(cs.Component):
    """A physical component on the Physical Architecture layer."""

    _xmltag = "ownedPhysicalComponents"

    functions = c.ProxyAccessor(
        PhysicalFunction,
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )
    realized_logical_components = c.ProxyAccessor(
        la.LogicalComponent,
        cs.ComponentRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )
    ports = c.ProxyAccessor(cs.PhysicalPort, aslist=c.ElementList)

    components: c.Accessor


@c.xtype_handler(XT_ARCH)
class PhysicalComponentPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedPhysicalComponentPkg"

    components = c.ProxyAccessor(PhysicalComponent, aslist=c.ElementList)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


class PhysicalArchitecture(crosslayer.BaseArchitectureLayer):
    """Provides access to the Physical Architecture layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        PhysicalComponent,
        attributes={"is_actor": False},
        rootelem=PhysicalComponentPkg,
    )
    root_function = c.ProxyAccessor(
        PhysicalComponent, rootelem=PhysicalFunctionPkg
    )

    component_package = c.ProxyAccessor(PhysicalComponentPkg)
    function_package = c.ProxyAccessor(PhysicalFunctionPkg)

    all_functions = c.ProxyAccessor(
        PhysicalFunction,
        aslist=c.ElementList,
        rootelem=PhysicalFunctionPkg,
        deep=True,
    )
    all_actors = property(
        lambda self: self._model.search(PhysicalComponent).by_is_actor(True)
    )
    all_components = c.ProxyAccessor(
        PhysicalComponent, aslist=c.ElementList, deep=True
    )

    all_physical_exchanges = c.ProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[PhysicalFunctionPkg, PhysicalFunction],
        deep=True,
    )
    all_physical_links = c.ProxyAccessor(
        cs.PhysicalLink, aslist=c.ElementList, deep=True
    )

    diagrams = diagram.DiagramAccessor(
        "Physical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )  # type: ignore[assignment]


c.set_accessor(
    la.LogicalComponent,
    "realizing_physical_components",
    c.CustomAccessor(
        PhysicalComponent,
        operator.attrgetter("_model.pa.all_components"),
        matchtransform=operator.attrgetter("realized_logical_components"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    la.LogicalFunction,
    "realizing_physical_functions",
    c.CustomAccessor(
        PhysicalFunction,
        operator.attrgetter("_model.pa.all_functions"),
        matchtransform=operator.attrgetter("realized_logical_functions"),
        aslist=c.ElementList,
    ),
)
c.set_self_references(
    (PhysicalComponent, "components"),
    (PhysicalComponentPkg, "packages"),
    (PhysicalFunction, "functions"),
    (PhysicalFunctionPkg, "packages"),
)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The decoration factories for svg elements."""
from __future__ import annotations

import collections.abc as cabc
import logging
import re
import typing as t

logger = logging.getLogger(__name__)

icon_size = 20
icon_padding = 2
feature_space = 24

function_ports = {"FIP", "FOP"}
directed_component_ports = {"CP_IN", "CP_OUT"}
component_ports = directed_component_ports | {"PP", "CP_INOUT", "CP_UNSET"}
all_ports = function_ports | component_ports
all_directed_ports = directed_component_ports | function_ports
start_aligned = {
    "LogicalComponent",
    "LogicalActor",
    "LogicalHumanActor",
    "LogicalHumanComponent",
}
only_icons = {"Requirement"}
# TODO: Instead of dirty-patching either rename FunctionalExchange in OA-Layer as
# ActivityExchange/OperationalExchange
DiagramClass = str
FaultyClass = str
PatchClass = str

needs_patch: dict[DiagramClass, dict[FaultyClass, PatchClass]] = {
    "Operational Entity Blank": {"FunctionalExchange": "OperationalExchange"}
}
always_top_label = {"Note", "Class", "Enumeration"}
needs_feature_line = {"Class", "Enumeration"}


class DecoFactories(t.Dict[str, cabc.Callable]):
    def __call__(self, func: cabc.Callable) -> cabc.Callable:
        symbol_name = re.sub(
            "(?:^|_)([a-z])",
            lambda m: m.group(1).capitalize(),
            func.__name__,
        )
        self[symbol_name] = func
        return func

    def __missing__(self, class_: str) -> cabc.Callable:
        logger.error("%s wasn't found in factories.", class_)
        assert "ErrorSymbol" in self
        return self["ErrorSymbol"]


deco_factories = DecoFactories()



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Core functionality shared by all PVMT submodules."""

from capellambse.loader.xmltools import AttributeProperty
from capellambse.loader.xmltools import XMLDictProxy as XMLDictProxy_

NAMESPACES = dict(xsi="http://www.w3.org/2001/XMLSchema-instance")


class Generic:
    """Base class for PropertyValues, Domains etc."""

    idx = AttributeProperty("xml_element", "id", writable=False)
    # TODO Investigate potential problems with renaming things
    name = AttributeProperty("xml_element", "name")
    sid = AttributeProperty("xml_element", "sid", optional=True)

    def __init__(self, xml_element, *args, **kwargs):
        super().__init__(*args, **kwargs)  # type: ignore[call-arg]
        self.xml_element = xml_element

    def __repr__(self):
        return '<pvmt.{} "{}"({})>'.format(
            type(self).__name__, self.name, self.idx
        )

    @classmethod
    def from_xml_element(cls, element):
        """Construct an object from the given XML element.

        This function is used to allow subclasses more control over how
        they are instantiated from existing XML elements, compared to
        creating them from scratch.
        """
        return cls(element)


class XMLDictProxy(Generic, XMLDictProxy_):  # pylint: disable=abstract-method
    """Facilitates usage of the :class:`capellambse.loader.XMLDictProxy`."""

    def __init__(self, xml_element, *args, **kwargs):
        # pylint: disable=arguments-out-of-order
        super().__init__(xml_element, xml_element, *args, **kwargs)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import markupsafe

__all__ = [
    "Accessor",
    "WritableAccessor",
    "ProxyAccessor",
    "AttrProxyAccessor",
    "AlternateAccessor",
    "ParentAccessor",
    "CustomAccessor",
    "AttributeMatcherAccessor",
    "SpecificationAccessor",
    "ReferenceSearchingAccessor",
    "ElementListCouplingMixin",
]

import abc
import collections.abc as cabc
import itertools
import operator
import typing as t

from lxml import etree

import capellambse
from capellambse import helpers

from . import XTYPE_HANDLERS, S, T, U, build_xtype, element

_NOT_SPECIFIED = object()
"Used to detect unspecified optional arguments"


class Accessor(t.Generic[T], metaclass=abc.ABCMeta):
    """Super class for all Accessor types."""

    __slots__ = (
        "__name__",
        "__objclass__",
    )

    __objclass__: type[t.Any]
    __name__: str

    @t.overload
    def __get__(self, obj: None, objtype=None) -> Accessor:
        ...

    @t.overload
    def __get__(self, obj, objtype=None) -> T | element.ElementList[T]:
        ...

    @abc.abstractmethod
    def __get__(self, obj, objtype=None):
        pass

    def __set__(self, obj: element.GenericElement, value: t.Any) -> None:
        raise TypeError("Cannot set this type of attribute")

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        self.__objclass__ = owner
        self.__name__ = name

    def __repr__(self) -> str:
        return f"<'{self.__objclass__.__name__}.{self.__name__}' {type(self).__name__}>"


class WritableAccessor(Accessor[T], metaclass=abc.ABCMeta):
    """An Accessor that also provides write support on lists it returns."""

    __slots__ = ()

    aslist: type[ElementListCouplingMixin] | None

    def __init__(
        self,
        *args: t.Any,
        aslist: type[element.ElementList] | None,
        **kw: t.Any,
    ) -> None:
        super().__init__(*args, **kw)  # type: ignore[call-arg]

        if aslist is not None:
            self.aslist = type(
                "Coupled" + aslist.__name__,
                (ElementListCouplingMixin, aslist),
                {"_accessor": self},
            )
            self.aslist.__module__ = __name__
        else:
            self.aslist = None

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        """Create and return a new element of type ``elmclass``.

        Parameters
        ----------
        elmlist
            The (coupled) ``ElementList`` to insert the new object into.
        xtype
            The ``xsi:type`` of the new object.
        elmclass
            The concrete ``GenericElement`` subclass to instantiate.
        kw
            Additional keyword arguments that are passed to the
            ``elmclass`` constructor.
        """
        raise TypeError("Cannot create objects")

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject,
    ) -> None:
        """Insert the ``value`` object into the model.

        The object must be inserted at an appropriate place, so that, if
        ``elmlist`` were to be created afresh, ``value`` would show at
        at index ``index``.
        """
        raise NotImplementedError("Objects cannot be inserted into this list")

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        """Delete the ``obj`` from the model."""
        raise NotImplementedError("Objects in this list cannot be deleted")

    def _match_xtype(
        self,
        type_1: str | object | None = _NOT_SPECIFIED,
        type_2: str | object = _NOT_SPECIFIED,
        /,
    ) -> tuple[type[T], str]:
        r"""Find the right class for the given ``xsi:type``\ (s)."""
        if type_1 is _NOT_SPECIFIED and type_2 is _NOT_SPECIFIED:
            elmclass = getattr(self, "elmclass", None)
            if elmclass not in (None, element.GenericElement):
                return elmclass
            raise TypeError("No object type specified")

        def match_xt(xtp: S, itr: cabc.Iterable[S]) -> S:
            matches: list[S] = []
            for i in itr:
                if (
                    xtp is i is None
                    or i is not None
                    and xtp in (i, i.split(":")[-1])
                ):
                    matches.append(i)
            if not matches:
                raise ValueError(f"Invalid or unknown xsi:type {xtp!r}")
            if len(matches) > 1:
                raise ValueError(
                    f"Ambiguous xsi:type {xtp!r}, please qualify: "
                    + ", ".join(repr(i) for i in matches)
                )
            return matches[0]

        if not isinstance(type_1, str):
            raise TypeError(
                f"Expected str as first type, got {type(type_1).__name__!r}"
            )
        candidate_classes: dict[str, type[T]]
        if type_2 is _NOT_SPECIFIED:
            candidate_classes = dict(
                itertools.chain.from_iterable(
                    i.items() for i in XTYPE_HANDLERS.values()
                )
            )
            objtype = type_1
        elif not isinstance(type_2, str):
            raise TypeError(
                f"Expected a str objtype, not {type(type_2).__name__}"
            )
        else:
            candidate_classes = XTYPE_HANDLERS[
                match_xt(type_1, XTYPE_HANDLERS)
            ]
            objtype = type_2

        objtype = match_xt(objtype, candidate_classes)
        return candidate_classes[objtype], objtype


class PhysicalAccessor(Accessor[T]):
    """Helper super class for accessors that work with real elements."""

    __slots__ = (
        "aslist",
        "class_",
        "list_extra_args",
        "xtypes",
    )

    aslist: type[element.ElementList] | None
    class_: type[T]
    xtypes: cabc.Set[str]

    def __init__(
        self,
        class_: type[T],
        xtypes: (
            str
            | type[element.ModelObject]
            | cabc.Iterable[str | type[element.ModelObject]]
            | None
        ) = None,
        *,
        aslist: type[element.ElementList[T]] | None = None,
        list_extra_args: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        if xtypes is None:
            self.xtypes = (
                {build_xtype(class_)}
                if class_ is not element.GenericElement
                else set()
            )
        elif isinstance(xtypes, type):
            assert issubclass(xtypes, element.GenericElement)
            self.xtypes = {build_xtype(xtypes)}
        elif isinstance(xtypes, str):
            self.xtypes = {xtypes}
        else:
            self.xtypes = {
                i if isinstance(i, str) else build_xtype(i) for i in xtypes
            }

        self.aslist = aslist
        self.class_ = class_
        self.list_extra_args = list_extra_args or {}

    def _guess_xtype(self) -> tuple[type[T], str]:
        """Try to guess the type of element that should be created."""
        if self.class_ is element.GenericElement or self.class_ is None:
            raise ValueError("Multiple object types that can be created")
        if not self.xtypes:
            raise ValueError("No matching `xsi:type` found")
        if len(self.xtypes) > 1:
            raise ValueError("Multiple matching `xsi:type`s")
        return self.class_, next(iter(self.xtypes))

    def _make_list(self, parent_obj, elements):
        if self.aslist is None:
            return no_list(self, parent_obj._model, elements, self.class_)
        return self.aslist(
            parent_obj._model,
            elements,
            self.class_,
            **self.list_extra_args,
        )


class ProxyAccessor(WritableAccessor[T], PhysicalAccessor[T]):
    """Creates proxy objects on the fly."""

    __slots__ = (
        "deep",
        "follow",
        "follow_abstract",
        "rootelem",
    )

    aslist: type[ElementListCouplingMixin] | None

    def __init__(
        self,
        class_: type[T],
        xtypes: str | type[T] | cabc.Iterable[str | type[T]] | None = None,
        *,
        aslist: type[element.ElementList] | None = None,
        deep: bool = False,
        follow: str | None = None,
        follow_abstract: bool = True,
        rootelem: (
            str
            | type[element.GenericElement]
            | cabc.Sequence[str | type[element.GenericElement]]
            | None
        ) = None,
        list_extra_args: dict[str, t.Any] | None = None,
    ):
        """Create a ProxyAccessor.

        Parameters
        ----------
        class_
            The proxy class.
        xtypes
            The ``xsi:type``(s) of the child element(s).  If None, then
            the constructed proxy will be passed the original element
            instead of a child.
        aslist
            If None, only a single element must match, which will be
            returned directly.  If not None, must be a subclass of
            :class:`elements.DecoupledElementList`.  It will be used to return a
            list of all matched objects.  Incompatible with ``xtypes =
            None``.
        deep
            True to search the entire XML subtree, False to only
            consider direct children.  Ignored if ``xtype = None``.
        follow
            If not None, must be the name of an attribute on the found
            elements.  That attribute contains a reference to the actual
            element that will be used instead.
        rootelem
            A ``/``-separated list of ``xsi:type``s that defines the
            path from the current object's element to the search root.
            If None, the current element will be used directly.
        """
        super().__init__(
            class_, xtypes, aslist=aslist, list_extra_args=list_extra_args
        )
        self.deep: bool = deep
        self.follow: str | None = follow
        self.follow_abstract: bool = follow_abstract
        if rootelem is None:
            self.rootelem: cabc.Sequence[str] = []
        elif isinstance(rootelem, str):
            self.rootelem = rootelem.split("/")
        elif isinstance(rootelem, type) and issubclass(
            rootelem, element.GenericElement
        ):
            self.rootelem = [build_xtype(rootelem)]
        else:
            self.rootelem = [
                i if isinstance(i, str) else build_xtype(i) for i in rootelem
            ]

    def __get__(self, obj, objtype=None):
        if obj is None:  # pragma: no cover
            return self

        elems = [self.__follow_attr(obj, e) for e in self.__getsubelems(obj)]
        elems = [x for x in elems if x is not None]
        return self._make_list(obj, elems)

    def __getsubelems(
        self, obj: element.ModelObject
    ) -> cabc.Iterator[etree._Element]:
        yielded_uuids = {None}

        if self.deep:
            it_func = obj._model._loader.iterdescendants_xt
        else:
            it_func = obj._model._loader.iterchildren_xt
        for elem in itertools.chain.from_iterable(
            it_func(i, *iter(self.xtypes)) for i in self.__findroots(obj)
        ):
            elemid = elem.get("id")
            if elemid in yielded_uuids:
                continue

            yielded_uuids.add(elemid)
            yield elem

    def __findroots(self, obj: element.ModelObject) -> list[etree._Element]:
        roots = [obj._element]
        for xtype in self.rootelem:
            roots = list(
                itertools.chain.from_iterable(
                    obj._model._loader.iterchildren_xt(i, xtype) for i in roots
                )
            )
        return roots

    def __follow_attr(
        self, obj: element.ModelObject, elem: etree._Element
    ) -> etree._Element:
        if self.follow:
            if self.follow in elem.attrib:
                elem = obj._model._loader[elem.attrib[self.follow]]
            else:
                return None
        return self.__follow_href(obj, elem)

    def __follow_href(
        self, obj: element.ModelObject, elem: etree._Element
    ) -> etree._Element:
        href = elem.get("href")
        if href:
            elem = obj._model._loader[href]
        if self.follow_abstract:
            abstype = elem.get("abstractType")
            if abstype is not None:
                elem = obj._model._loader[abstype]
        return elem

    def _make_list(self, parent_obj, elements):
        if self.aslist is None:
            return no_list(self, parent_obj._model, elements, self.class_)
        return self.aslist(
            parent_obj._model,
            elements,
            self.class_,
            parent=parent_obj,
            **self.list_extra_args,
        )

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        if self.deep or self.follow or self.rootelem:
            raise TypeError("Cannot create objects here")

        if type_hints:
            elmclass, xtype = self._match_xtype(*type_hints)
        else:
            elmclass, xtype = self._guess_xtype()

        assert elmclass is not None
        assert isinstance(elmlist._parent, element.GenericElement)
        assert issubclass(elmclass, element.GenericElement)

        parent = elmlist._parent._element
        with elmlist._model._loader.new_uuid(parent) as obj_id:
            obj = elmclass(
                elmlist._model, parent, **kw, xtype=xtype, uuid=obj_id
            )
        return obj  # type: ignore[return-value]

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject,
    ) -> None:
        if value._model is not elmlist._model:
            raise ValueError("Cannot move elements between models")
        try:
            indexof = elmlist._parent._element.index
            if index > 0:
                parent_index = indexof(elmlist._elements[index - 1]) + 1
            elif index < -1:
                parent_index = indexof(elmlist._elements[index + 1]) - 1
            else:
                parent_index = index
        except ValueError:
            # FIXME: self._parent is part of ElementListCouplingMixin
            parent_index = len(self._parent)  # type: ignore[attr-defined]
        elmlist._parent._element.insert(parent_index, value._element)
        elmlist._model._loader.idcache_index(value._element)

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        assert obj._model is elmlist._model
        elmlist._model._loader.idcache_remove(obj._element)
        elmlist._parent._element.remove(obj._element)


class AttrProxyAccessor(PhysicalAccessor[T]):
    """Provides access to elements that are linked in an attribute."""

    __slots__ = ("attr",)

    def __init__(
        self,
        class_: type[T],
        attr: str,
        *,
        aslist: type[element.ElementList] | None = None,
    ):
        """Create an AttrProxyAccessor.

        Parameters
        ----------
        class_
            The proxy class.
        attr
            The element attribute to handle.
        aslist
            If None, the attribute contains at most one element
            reference, and either None or the constructed proxy will be
            returned.  If not None, must be a subclass of
            :class:`elements.DecoupledElementList`.  It will be used to return a
            list of all matched objects.  Incompatible with ``xtypes =
            None``.
        """
        super().__init__(class_, aslist=aslist)
        self.attr = attr

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        elems = []
        next_xtype: str | None = None
        for elemref in obj._element.get(self.attr, "").split():
            if "#" in elemref:
                elem = obj._model._loader[elemref]
                if next_xtype is not None:
                    this_xtype = helpers.xtype_of(elem)
                    if this_xtype != next_xtype:  # pragma: no cover
                        raise RuntimeError(
                            "Broken XML: Expected xsi:type {!r}, got {!r}".format(
                                next_xtype, this_xtype
                            )
                        )
                elems.append(elem)
                next_xtype = None
            elif next_xtype is not None:  # pragma: no cover
                raise RuntimeError(
                    "Broken XML: Expected element reference, got xsi:type {!r}".format(
                        elemref
                    )
                )
            else:
                next_xtype = elemref

        return self._make_list(obj, elems)

    def __set__(
        self,
        obj: element.GenericElement,
        values: T | cabc.Iterable[T],
    ) -> None:
        if not isinstance(values, cabc.Iterable):
            values = (values,)
        elif self.aslist is None:
            raise TypeError(
                f"{self.__objclass__.__name__}.{self.__name__}"
                " requires a single item, not an iterable"
            )

        assert isinstance(values, cabc.Iterable)
        parts: list[str] = []
        for value in values:
            if not value._model is obj._model:
                raise ValueError("Cannot set elements from different models")
            link = obj._model._loader.create_link(obj._element, value._element)
            parts.append(link)
        obj._element.set(self.attr, " ".join(parts))


class AlternateAccessor(PhysicalAccessor[T]):
    """Provides access to an "alternate" form of the object."""

    __slots__ = ()

    def __init__(
        self,
        class_: type[T],
    ):
        super().__init__(class_)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self
        return self.class_.from_model(obj._model, obj._element)


class ParentAccessor(PhysicalAccessor[T]):
    """Accesses the parent XML element."""

    __slots__ = ()

    def __init__(
        self,
        class_: type[T],
    ):
        super().__init__(class_)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self
        return self.class_.from_model(
            obj._model,
            obj._element.getparent(),
        )


class CustomAccessor(PhysicalAccessor[T]):
    """Customizable alternative to the ProxyAccessor."""

    __slots__ = (
        "elmfinders",
        "elmmatcher",
        "matchtransform",
    )

    def __init__(
        self,
        class_: type[T],
        *elmfinders: cabc.Callable[[element.GenericElement], cabc.Iterable[T]],
        elmmatcher: cabc.Callable[
            [U, element.GenericElement], bool
        ] = operator.contains,  # type: ignore[assignment]
        matchtransform: cabc.Callable[[T], U] = (
            lambda e: e  # type: ignore[assignment,return-value]
        ),
        aslist: type[element.ElementList] | None = None,
    ) -> None:
        """Create a CustomAccessor.

        Parameters
        ----------
        class_
            The target subclass of ``GenericElement``
        elmfinders
            Functions that are called on the current element.  Each
            returns an iterable of possible targets.
        elmmatcher
            Function that is called with the transformed target element
            and the current element to determine if the untransformed
            target should be accepted.
        matchtransform
            Function that transforms a target so that it can be used by
            the matcher function.
        """
        super().__init__(class_, aslist=aslist)
        self.elmfinders = elmfinders
        self.elmmatcher = elmmatcher
        self.matchtransform = matchtransform

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        elms = itertools.chain.from_iterable(f(obj) for f in self.elmfinders)
        matches = [
            e._element
            for e in elms
            if self.elmmatcher(self.matchtransform(e), obj)
        ]
        return self._make_list(obj, matches)


class AttributeMatcherAccessor(ProxyAccessor[T]):
    __slots__ = (
        "_AttributeMatcherAccessor__aslist",
        "attributes",
    )

    def __init__(
        self,
        class_: type[T],
        xtypes: str | type[T] | cabc.Iterable[str | type[T]] | None = None,
        *,
        aslist: type[element.ElementList] | None = None,
        attributes: dict[str, t.Any],
        **kwargs,
    ) -> None:
        super().__init__(
            class_, xtypes, aslist=element.MixedElementList, **kwargs
        )
        self.__aslist = aslist
        self.attributes = attributes

    def __get__(self, obj, objtype=None):
        if obj is None:  # pragma: no cover
            return self

        elements = super().__get__(obj, objtype)
        matches = []
        assert isinstance(elements, cabc.Iterable)
        for elm in elements:
            try:
                if all(
                    getattr(elm, k) == v for k, v in self.attributes.items()
                ):
                    matches.append(elm._element)
            except AttributeError:
                pass

        if self.__aslist is None:
            return no_list(self, obj._model, matches, self.class_)
        return self.__aslist(obj._model, matches, self.class_)


class _Specification(t.MutableMapping[str, str], element.ModelObject):
    _aliases = {"LinkedText": "capella:linkedText"}
    _linked_text = frozenset({"capella:linkedText"})

    def __init__(
        self, model: capellambse.MelodyModel, elm: etree._Element
    ) -> None:
        self._model = model
        self._element = elm

    def __delitem__(self, k: str) -> None:
        k = self._aliases.get(k, k)
        i, lang_elem = self._index_of(k)
        body_elem = self._body_at(k, i)
        self._element.remove(lang_elem)
        self._element.remove(body_elem)

    def __getitem__(self, k: str) -> str:
        k = self._aliases.get(k, k)
        i, _ = self._index_of(k)
        v = self._body_at(k, i).text or ""
        if k in self._linked_text:
            v = helpers.unescape_linked_text(self._model._loader, v)
        return v

    def __iter__(self) -> cabc.Iterator[str]:
        for i in self._element.iterchildren("languages"):
            yield i.text or ""

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __setitem__(self, k: str, v: str) -> None:
        k = self._aliases.get(k, k)
        if k in self._linked_text:
            v = helpers.escape_linked_text(self._model._loader, v)
        try:
            i, lang = self._index_of(k)
        except KeyError:
            self._element.append(body := self._element.makeelement("bodies"))
            self._element.append(
                lang := self._element.makeelement("languages")
            )
            body.text = v
            lang.text = k
        else:
            body = self._body_at(k, i)
            body.text = v

    def _index_of(self, k: str) -> tuple[int, etree._Element]:
        for i, elm in enumerate(self._element.iterchildren("languages")):
            if elm.text == k:
                return i, elm

        raise KeyError(k)

    def _body_at(self, k: str, i: int) -> etree._Element:
        try:
            return next(
                itertools.islice(
                    self._element.iterchildren("bodies"), i, i + 1
                )
            )
        except StopIteration:
            raise KeyError(k) from None

    def __html__(self) -> str:
        return markupsafe.escape(self.__str__())

    def __str__(self) -> str:  # pragma: no cover
        return next(iter(self.values()))

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{type(self).__name__} at 0x{id(self):016X} {list(self)!r}>"

    @classmethod
    def from_model(
        cls, _1: capellambse.MelodyModel, _2: t.Any
    ) -> element.ModelObject:
        """Specifications can not be instantiated"""
        raise RuntimeError("Can not create a specification from a model")


class SpecificationAccessor(Accessor[_Specification]):
    """Provides access to linked specifications."""

    __slots__ = ()

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        try:
            spec_elm = next(obj._element.iterchildren("ownedSpecification"))
        except StopIteration:
            raise AttributeError("No specification found") from None

        return _Specification(obj._model, spec_elm)


class ReferenceSearchingAccessor(PhysicalAccessor[T]):
    __slots__ = ("attrs",)

    attrs: tuple[str, ...]

    def __init__(
        self,
        class_: type[T],
        *attrs: str,
        aslist: type[element.ElementList] | None = None,
    ) -> None:
        super().__init__(class_, aslist=aslist)
        self.attrs = attrs

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        matches: list[etree._Element] = []
        for candidate in obj._model.search(self.class_):
            for attr in self.attrs:
                try:
                    value = getattr(candidate, attr)
                except AttributeError:
                    continue
                if (
                    isinstance(value, element.ElementList)
                    and obj in value
                    or isinstance(value, element.GenericElement)
                    and obj == value
                ):
                    matches.append(candidate._element)
                    break
        return self._make_list(obj, matches)


def no_list(
    desc: Accessor,
    model: capellambse.MelodyModel,
    elems: cabc.Sequence[etree._Element],
    class_: type[T],
) -> element.ModelObject | None:
    """Return a single element or None instead of a list of elements.

    Parameters
    ----------
    desc
        The descriptor that called this function
    model
        The ``MelodyModel`` instance
    elems
        List of elements that was matched
    class_
        The ``GenericElement`` subclass to instantiate
    """
    if not elems:  # pragma: no cover
        return None
    if len(elems) > 1:  # pragma: no cover
        raise RuntimeError(
            "Expected 1 object for {}.{}, got {}".format(
                desc.__objclass__.__name__, desc.__name__, len(elems)
            )
        )
    return class_.from_model(model, elems[0])


class ElementListCouplingMixin(element.ElementList[T], t.Generic[T]):
    """Couples an ElementList with an Accessor to enable write support.

    This class is meant to be subclassed further, where the subclass has
    both this class and the originally intended one as base classes (but
    no other ones, i.e. there must be exactly two bases).  The Accessor
    then inserts itself as the ``_accessor`` class variable on the new
    subclass.  This allows the mixed-in methods to delegate actual model
    modifications to the Accessor.
    """

    _accessor: t.ClassVar[WritableAccessor[T]]

    def __init__(
        self, *args: t.Any, parent: element.ModelObject, **kw: t.Any
    ) -> None:
        assert type(self)._accessor
        assert isinstance(parent, element.GenericElement)

        super().__init__(*args, **kw)
        self._parent = parent

    @t.overload
    def __setitem__(self, index: int, value: T) -> None:
        ...

    @t.overload
    def __setitem__(self, index: slice, value: cabc.Iterable[T]) -> None:
        ...

    def __setitem__(
        self, index: int | slice, value: T | cabc.Iterable[T]
    ) -> None:
        assert self._parent is not None
        del self[index]
        if isinstance(index, slice):
            assert isinstance(value, cabc.Iterable)
            for i, elm in enumerate(value, start=index.start):
                self.insert(i, elm)
        else:
            assert not isinstance(value, cabc.Iterable)
            self.insert(index, value)

    def __delitem__(self, index: int | slice) -> None:
        assert self._parent is not None
        if not isinstance(index, slice):
            index = slice(index, index + 1)
        for obj in self[index]:
            accessor = type(self)._accessor
            assert isinstance(accessor, WritableAccessor)
            accessor.delete(self, obj)
        super().__delitem__(index)

    def _newlist_type(self) -> type[element.ElementList[T]]:
        assert len(type(self).__bases__) == 2
        assert type(self).__bases__[0] is ElementListCouplingMixin
        return type(self).__bases__[1]

    def create(self, *args: str | None, **kw: t.Any) -> T:
        """Make a new model object (instance of GenericElement).

        Instead of specifying the full ``xsi:type`` including the
        namespace, you can also pass in just the part after the ``:``
        separator.  If this is unambiguous, the appropriate
        layer-specific type will be selected automatically.

        This method can be called with or without the ``layertype``
        argument.  If a layertype is not given, all layers will be tried
        to find an appropriate ``xsi:type`` handler.  Note that setting
        the layertype to ``None`` explicitly is different from not
        specifying it at all; ``None`` tries only the "Transverse
        modelling" type elements.

        Parameters
        ----------
        layertype
            The ``xsi:type`` of the architectural layer on which this
            element will eventually live (see above).
        objtype
            The ``xsi:type`` of the object to create.
        """
        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, WritableAccessor)
        newobj = acc.create(self, *args, **kw)
        self._newlist_type().insert(self, len(self), newobj)
        return newobj

    def delete_all(self, **kw: t.Any) -> None:
        """Delete all matching objects from the model."""
        indices: list[int] = []
        for i, obj in enumerate(self):
            if all(getattr(obj, k) == v for k, v in kw.items()):
                indices.append(i)

        for index in reversed(indices):
            del self[index]

    def insert(self, index: int, value: T) -> None:
        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, WritableAccessor)
        acc.insert(self, index, value)
        self._newlist_type().insert(self, index, value)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser entry point for visual elements.

Visual elements, contrary to semantic ones, do not exist in the melody
files at all.  They are used for things such as Notes and inter-diagram
hyperlinks.
"""
from __future__ import annotations

import collections.abc as cabc
import logging

from capellambse import aird

from . import _common as c
from . import _edge_factories, _styling

LOGGER = logging.getLogger(__name__)


def from_xml(ebd: c.ElementBuilder) -> aird.DiagramElement:
    """Deserialize a visual element."""
    el_type = ebd.data_element.attrib[c.ATT_XMT].split(":")[-1]
    if el_type not in VISUAL_TYPES:
        LOGGER.error("Unknown visual element type, skipping: %r", el_type)
        raise c.SkipObject()
    factory = VISUAL_TYPES[el_type]
    if factory is None:
        raise c.SkipObject()
    return factory(ebd)


def connector_factory(ebd: c.ElementBuilder) -> aird.Edge:
    """Create a Connector.

    A Connector is an edge that connects Notes to arbitrary elements.
    """
    styleclass = "Connector"
    bendpoints, source, target = _edge_factories.extract_bendpoints(ebd)
    styleoverrides = _styling.apply_style_overrides(
        ebd.target_diagram.styleclass, styleclass, ebd.data_element
    )
    edge = aird.Edge(
        bendpoints,
        uuid=ebd.data_element.attrib[c.ATT_XMID],
        styleclass=styleclass,
        # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
        styleoverrides=styleoverrides,  # type: ignore[arg-type]
    )
    assert edge.uuid is not None

    if isinstance(target, aird.Box):
        _edge_factories.snaptarget(edge.points, -1, -2, target)
        target.add_context(edge.uuid)
    if isinstance(source, aird.Box):
        _edge_factories.snaptarget(edge.points, 0, 1, source)
        source.add_context(edge.uuid)

    return edge


def shape_factory(ebd: c.ElementBuilder) -> aird.Box:
    """Create a Shape."""
    assert ebd.target_diagram.styleclass is not None

    uid = ebd.data_element.attrib[c.ATT_XMID]
    label = ebd.data_element.attrib["description"]
    parent = ebd.data_element.getparent()
    while parent.tag == "children":
        parent_uid = parent.attrib.get("element") or parent.attrib.get(
            c.ATT_XMID
        )
        try:
            parent = ebd.target_diagram[parent_uid]
        except KeyError:
            parent = parent.getparent()
        else:
            refpos = parent.pos
            break
    else:
        parent = None
        refpos = aird.Vector2D(0, 0)

    try:
        layout = next(ebd.data_element.iterchildren("layoutConstraint"))
    except StopIteration:
        raise ValueError(
            "No layoutConstraint found for element {uid}"
        ) from None

    pos = refpos + (
        int(layout.attrib.get("x", "0")),
        int(layout.attrib.get("y", "0")),
    )
    size = aird.Vector2D(
        int(layout.attrib.get("width", "0")),
        int(layout.attrib.get("height", "0")),
    )
    styleclass = ebd.data_element.attrib["type"]
    styleoverrides = _styling.apply_visualelement_styles(
        ebd.target_diagram.styleclass, f"Box.{styleclass}", ebd.data_element
    )
    return aird.Box(
        pos,
        size,
        label=label,
        uuid=uid,
        parent=parent,
        styleclass=styleclass,
        # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
        styleoverrides=styleoverrides,  # type: ignore[arg-type]
    )


VISUAL_TYPES: dict[
    str, cabc.Callable[[c.ElementBuilder], aird.DiagramElement]
] = {
    "BasicDecorationNode": c.SkipObject.raise_,
    "Connector": connector_factory,
    # Nodes are actually semantic elements.  If one got through to here,
    # it's an internal node, not an element root.
    "Node": c.SkipObject.raise_,
    "Shape": shape_factory,
}



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The MelodyLoader loads and provides access to a Capella model.

It is using LXML internally to efficiently parse and navigate through
the Capella-generated XML files.  For more information about LXML, see
the `LXML Documentation`_.

.. _LXML Documentation: https://lxml.de/
"""

from .core import MelodyLoader
from .filehandler import FileHandler, get_filehandler
from .modelinfo import ModelInfo



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .. import common as c


@c.xtype_handler(None)
class Constraint(c.GenericElement):
    """A constraint."""

    _xmltag = "ownedConstraints"

    constrained_elements = c.AttrProxyAccessor(
        c.GenericElement,
        "constrainedElements",
        aslist=c.MixedElementList,
    )

    specification = c.SpecificationAccessor()


@c.xtype_handler(None)
class Generalization(c.GenericElement):
    """A Generalization"""

    _xmltag = "ownedGeneralizations"


c.set_accessor(
    c.GenericElement,
    "constraints",
    c.ProxyAccessor(Constraint, aslist=c.ElementList),
)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sphinx extension for python-capella-mbse.

This extension is be used to display diagrams in Sphinx-generated
documentation.

Usage
-----

To render a diagram in your documentation, simply define a diagram directive.

.. code::

    .. diagram:: [CDB] BaseLayer ORM
        :alt: Base layer diagram
        :height: 480
        :width: 640
        :align: center -- can be left/right/center

The options are optional.

Configuration
-------------

To enable this extension, add it to your list of extensions in Sphinx'
``conf.py``.

.. code:: python

    # conf.py

    extensions = [
        ...,
        'capellambse.sphinx',
    ]

The following configuration variables are understood by this extension:

*   ``capellambse_model``: Path to the Capella model.

    Set this variable to the root ``.aird`` file of the model you want
    to use in the documentation.  The path must be relative to Sphinx'
    current working directory, which should be the directory containing
    the ``conf.py`` file.

Known limitations
-----------------

*   The extension currently does not detect changes to the model, nor
    does it track which source files are using it.  This means that,
    after changing the model, you need to force a full rebuild of all
    pages by passing ``--fresh-env`` to Sphinx' build command.
"""
from __future__ import annotations

import pathlib
import typing as t

import sphinx.util.docutils
from docutils import nodes
from docutils.parsers import rst

import capellambse

if t.TYPE_CHECKING:
    import sphinx.application
    import sphinx.environment


def setup(app: sphinx.application.Sphinx) -> dict[str, t.Any]:
    """Called by Sphinx to set up the extension."""
    app.add_config_value(
        "capellambse_model", "../model/Documentation.aird", "html"
    )
    app.add_directive("diagram", DiagramDirective)

    app.connect("env-before-read-docs", load_model)
    app.connect("env-updated", unload_model)

    return {
        "version": capellambse.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def load_model(
    app: sphinx.application.Sphinx,
    env: sphinx.environment.BuildEnvironment,
    *_: t.Any,
) -> None:
    """Load the model."""
    if app.confdir is None:
        raise ValueError("Cannot load model: No confdir defined for Sphinx")

    env.capellambse_loaded_model = capellambse.MelodyModel(  # type: ignore[attr-defined]
        pathlib.Path(app.confdir, app.config.capellambse_model)
    )


def unload_model(_: t.Any, env: sphinx.environment.BuildEnvironment) -> None:
    """Unload the model.

    The LXML tree cannot be pickled together with the rest of the build
    environment.  This means we need to unload the model again after
    processing the source files.
    """
    if hasattr(env, "capellambse_loaded_model"):
        del env.capellambse_loaded_model  # type: ignore[attr-defined]


class DiagramDirective(sphinx.util.docutils.SphinxDirective):
    """The "``.. diagram::``" reST directive."""

    has_content = ...
    required_arguments = 1
    final_argument_whitespace = True
    option_spec = {
        "alt": rst.directives.unchanged,  # type: ignore[attr-defined]
        "height": rst.directives.nonnegative_int,  # type: ignore[attr-defined]
        "width": rst.directives.nonnegative_int,  # type: ignore[attr-defined]
        "align": lambda arg: rst.directives.choice(  # type: ignore[attr-defined]
            arg, ("left", "center", "right")
        ),
    }

    def run(self) -> list[nodes.Node]:
        name = self.arguments[0]
        if not hasattr(self.env, "capellambse_loaded_model"):
            raise self.error(
                f"Cannot show diagram {name!r}: No model configured"
            )

        model = self.env.capellambse_loaded_model  # type: ignore[attr-defined]
        try:
            diagram = model.diagrams.by_name(name)
        except KeyError as error:
            raise self.error(
                f"Cannot find diagram {name!r} in the configured model"
            ) from error

        uri = diagram.as_datauri_svg
        return [
            nodes.image(rawsource=self.block_text, uri=uri, **self.options)
        ]



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Two dimensional vector calculation utility."""
# pylint: disable=unsubscriptable-object, not-an-iterable  # false-positives
from __future__ import annotations

__all__ = ["Vec2Element", "Vec2List", "Vec2Property", "Vec2ish", "Vector2D"]

import collections.abc as cabc
import math
import operator
import typing as t

_T = t.TypeVar("_T")

Vec2Element = t.Union[float, int]
Vec2ish = t.Tuple[Vec2Element, Vec2Element]


class Vector2D(t.NamedTuple):
    """A vector in 2-dimensional space."""

    x: Vec2Element = 0
    y: Vec2Element = 0

    def __add__(self, other: Vec2ish) -> Vector2D:  # type: ignore
        return self.__map2(operator.add, other)

    def __radd__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.add, other, True)

    def __sub__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.sub, other)

    def __rsub__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.sub, other, True)

    @t.overload  # type: ignore
    def __mul__(self, other: Vec2ish) -> Vec2Element:
        ...

    @t.overload
    def __mul__(self, other: Vec2Element) -> Vector2D:
        ...

    def __mul__(self, other: Vec2Element | Vec2ish) -> Vector2D | Vec2Element:
        result = self.__map2(operator.mul, other)
        if result is NotImplemented:
            return self.__map(operator.mul, other)
        return sum(result)

    @t.overload
    def __rmul__(self, other: Vec2ish) -> Vec2Element:
        ...

    @t.overload
    def __rmul__(self, other: Vec2Element) -> Vector2D:
        ...

    def __rmul__(self, other: Vec2Element | Vec2ish) -> Vector2D | Vec2Element:
        result = self.__map2(operator.mul, other, True)
        if result is NotImplemented:
            return self.__map(operator.mul, other, True)
        return sum(result)

    def __matmul__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.mul, other)

    def __rmatmul__(self, other: Vec2ish) -> Vector2D:
        return self.__map2(operator.mul, other, True)

    def __truediv__(self, other: Vec2Element) -> Vector2D:
        return self.__map(operator.truediv, other)

    def __rtruediv__(self, other: Vec2Element) -> Vector2D:
        return self.__map(operator.truediv, other, True)

    def __floordiv__(self, other: Vec2Element) -> Vector2D:
        result = self.__map(operator.floordiv, other)
        if result is NotImplemented:  # pragma: no cover
            return result
        return type(self)(int(result.x), int(result.y))

    def __rfloordiv__(self, other: Vec2Element) -> Vector2D:
        result = self.__map(operator.floordiv, other, True)
        if result is NotImplemented:  # pragma: no cover
            return result
        return type(self)(int(result.x), int(result.y))

    def __abs__(self) -> Vector2D:
        return type(self)(abs(self[0]), abs(self[1]))

    def __str__(self) -> str:  # pragma: no cover
        return f"({self[0]}, {self[1]})"

    @property
    def sqlength(self) -> float:
        """Calculate the squared length of this vector."""
        return self[0] ** 2 + self[1] ** 2

    @property
    def length(self) -> float:
        """Calculate the length of this vector."""
        return math.sqrt(self.sqlength)

    @property
    def normalized(self) -> Vector2D:
        """Create a unit Vector2D with the same direction as this one.

        Raises
        ------
        ZeroDivisionError
            if this Vector2D has zero length
        """
        length = self.length
        return Vector2D(*(i / length for i in self))

    def closestaxis(self) -> Vector2D:
        """Determine the axis closest to this Vector2D."""
        horizontal = abs(self[0]) >= abs(self[1])
        return Vector2D(*((-1, 1)[i >= 0] for i in self)) @ (
            horizontal,
            not horizontal,
        )

    def angleto(self, other: Vec2ish) -> float:
        """Calculate the angle to ``other``.

        This method calculates the angle this vector needs to be rotated
        by in order to have the same direction as ``other``, in radians.
        """
        angle = math.atan2(other[1], other[0]) - math.atan2(self[1], self[0])
        # Make sure we get the shortest rotation possible
        if angle > math.pi:  # pragma: no cover
            angle -= 2 * math.pi
        elif angle < -math.pi:  # pragma: no cover
            angle += 2 * math.pi
        return angle

    def rotatedby(self, theta: Vec2Element) -> Vector2D:
        """Rotate this Vector2D by ``theta`` radians."""
        cos_t = math.cos(theta)
        sin_t = math.sin(theta)

        return Vector2D(
            self[0] * cos_t - self[1] * sin_t,
            self[0] * sin_t + self[1] * cos_t,
        )

    def boxsnap(
        self, corner1: Vec2ish, corner2: Vec2ish, dirvec: Vec2ish = (0, 0)
    ) -> Vector2D:
        """Snap this vector to the side of a box and return the result.

        Parameters
        ----------
        corner1
            A Vector2D describing the first corner of the target box.
        corner2
            A Vector2D describing the second corner of the target box.
        dirvec
            A Vector2D pointing in the direction to snap towards.
        """
        minx = min(corner1[0], corner2[0])
        miny = min(corner1[1], corner2[1])
        maxx = max(corner1[0], corner2[0])
        maxy = max(corner1[1], corner2[1])

        if abs(dirvec[0]) >= abs(dirvec[1]):
            # Snap horizontally
            return Vector2D(
                (minx, maxx)[dirvec[0] > 0],
                min(maxy - 1, max(miny + 1, self[1])),
            )
        # Snap vertically
        return Vector2D(
            min(maxx - 1, max(minx + 1, self[0])), (miny, maxy)[dirvec[1] > 0]
        )

    def __map(
        self,
        func: cabc.Callable[[Vec2Element, Vec2Element], Vec2Element],
        other: Vec2Element | Vec2ish,
        reflected: bool = False,
    ) -> Vector2D:
        if not isinstance(other, (int, float)):  # pragma: no cover
            return NotImplemented
        if reflected:
            return type(self)(func(other, self[0]), func(other, self[1]))
        return type(self)(func(self[0], other), func(self[1], other))

    def __map2(
        self,
        func: cabc.Callable[[Vec2Element, Vec2Element], Vec2Element],
        other: Vec2Element | Vec2ish,
        reflected: bool = False,
    ) -> Vector2D:
        if isinstance(other, (int, float)):  # pragma: no cover
            return NotImplemented
        if not len(other) == 2:  # pragma: no cover
            raise ValueError("Length of 'other' must be 2")
        if reflected:
            return type(self)(func(other[0], self[0]), func(other[1], self[1]))
        return type(self)(func(self[0], other[0]), func(self[1], other[1]))


class Vec2Property:
    """A property that automatically converts 2-tuples into Vector2D."""

    __slots__ = ("default", "name", "__objclass__")
    default: Vector2D | None
    name: str | None
    __objclass__: type[t.Any]

    def __init__(self, default: Vec2ish | None = None):
        if default is None or isinstance(default, Vector2D):
            self.default = default
        else:
            self.default = Vector2D(*default)

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> Vec2Property:
        ...

    @t.overload
    def __get__(
        self, obj: t.Any, objtype: type[t.Any] | None = ...
    ) -> Vector2D:
        ...

    def __get__(
        self, obj: t.Any | None, objtype: type[t.Any] | None = None
    ) -> Vec2Property | Vector2D:
        if obj is None:
            return self
        if self.name is None:
            raise RuntimeError("This property does not have a name yet")
        try:
            return getattr(obj, f"_{type(self).__name__}__{self.name}")
        except AttributeError:
            if self.default:
                return self.default
            raise

    def __set__(self, obj: t.Any, value: Vec2ish) -> None:
        if self.name is None:
            raise RuntimeError("This property does not have a name yet")
        if not isinstance(value, Vector2D):
            value = Vector2D(*value)
        setattr(obj, f"_{type(self).__name__}__{self.name}", value)

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        self.__objclass__ = owner
        self.name = name


class Vec2List(t.MutableSequence[Vector2D]):
    """A list that automatically converts its elements into Vector2D."""

    def __init__(self, values: cabc.Iterable[Vec2ish]):
        self.__list: list[Vector2D] = []
        self.extend(values)

    def __len__(self) -> int:
        return len(self.__list)

    @t.overload
    def __getitem__(self, index: int) -> Vector2D:
        ...

    @t.overload
    def __getitem__(self, index: slice) -> cabc.MutableSequence[Vector2D]:
        ...

    def __getitem__(
        self, index: int | slice
    ) -> Vector2D | cabc.Sequence[Vector2D]:
        return self.__list[index]

    @t.overload
    def __setitem__(
        self,
        index: int,
        value: Vec2ish,
    ) -> None:
        ...

    @t.overload
    def __setitem__(
        self,
        index: slice,
        value: cabc.Iterable[Vec2ish],
    ) -> None:
        ...

    def __setitem__(
        self,
        index: int | slice,
        value: Vec2ish | cabc.Iterable[Vec2ish],
    ) -> None:
        if isinstance(index, slice):
            assert not isinstance(value, Vector2D)
            value = t.cast(cabc.Iterable[Vec2ish], value)
            self.__list[index] = (self.__cast(v) for v in value)
        else:
            assert isinstance(value, Vector2D)
            self.__list[index] = self.__cast(value)

    def __delitem__(self, index: int | slice) -> None:
        del self.__list[index]

    def append(self, value: Vec2ish) -> None:
        self.__list.append(self.__cast(value))

    def copy(self) -> Vec2List:
        """Create a copy of this Vec2List."""
        return Vec2List(self)

    def extend(self, values: cabc.Iterable[Vec2ish]) -> None:
        for i in values:
            self.append(i)

    def insert(self, index: int, value: Vector2D) -> None:
        self.__list.insert(index, self.__cast(value))

    @staticmethod
    def __cast(element: Vec2ish) -> Vector2D:
        if not isinstance(element, Vector2D):
            element = Vector2D(*element)
        return element



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An Eclipse-like XML serializer.

The libxml2 XML serializer produces very different output from the one
used by Capella.  This causes a file saved by libxml2 to look vastly
different, even though semantically nothing might have changed at all.
This module implements a serializer which produces output like Capella
does.
"""
from __future__ import annotations

import collections.abc as cabc
import contextlib
import html.entities
import io
import os
import re
import typing as t

import lxml.etree

INDENT = b"  "
LINESEP = os.linesep.encode("ascii")
LINE_LENGTH = 80

ESCAPE_CHARS = r"[\x00-\x1F\x7F{}]"
P_ESCAPE_TEXT = re.compile(ESCAPE_CHARS.format('"&<'))
P_ESCAPE_COMMENTS = re.compile(ESCAPE_CHARS.format(">"))
P_NAME = re.compile(r"^(?:\{([^}]*)\})?(.+)$")

ALWAYS_EXPANDED_TAGS = frozenset({"bodies"})


@t.runtime_checkable
class _HasWrite(t.Protocol):
    def write(self, chunk: bytes) -> int:
        ...


def to_string(tree: lxml.etree._Element, /) -> str:
    """Serialize an XML tree as a ``str``.

    No XML processing instruction will be inserted at the start of the
    document.

    Arguments
    ---------
    tree
        The XML tree to serialize.

    Returns
    -------
    str
        The serialized XML.
    """
    payload = serialize(tree, encoding="utf-8", errors="surrogateescape")
    return payload.decode("utf-8", errors="surrogateescape")


def to_bytes(
    tree: lxml.etree._Element,
    /,
    *,
    encoding: str = "utf-8",
    errors: str = "strict",
    declare_encoding: bool = True,
) -> bytes:
    """Serialize an XML tree as a ``str``.

    At the start of the document, an XML processing instruction will be
    inserted declaring the used encoding.  Pass
    ``declare_encoding=False`` to inhibit this behavior.

    Arguments
    ---------
    tree
        The XML tree to serialize.
    encoding
        The encoding to use.  An XML processing instruction will be
        inserted which declares the used encoding.
    errors
        How to handle errors during encoding.

    Returns
    -------
    bytes
        The serialized XML, encoded using ``encoding``.
    """
    if declare_encoding:
        declaration = _declare(encoding)
    else:
        declaration = b""
    return declaration + serialize(tree, encoding=encoding, errors=errors)


def write(
    tree: lxml.etree._Element,
    /,
    file: _HasWrite | os.PathLike | str | bytes,
    *,
    encoding: str = "utf-8",
    errors: str = "strict",
    line_length: float | int = LINE_LENGTH,
) -> None:
    """Write the XML tree to ``file``.

    Parameters
    ----------
    tree
        The XML tree to serialize.
    file
        An open file or a PathLike to write the XML into.
    encoding
        The file encoding to use when opening a file.
    errors
        Set the encoding error handling behavior of newly opened files.
    """
    ctx: t.ContextManager[_HasWrite]
    if isinstance(file, _HasWrite):
        ctx = contextlib.nullcontext(file)
    else:
        ctx = open(file, "wb")

    payload = serialize(
        tree, encoding=encoding, errors=errors, line_length=line_length
    )
    with ctx as f:
        f.write(_declare(encoding))
        f.write(payload)


def serialize(
    tree: lxml.etree._Element | lxml.etree._ElementTree,
    /,
    *,
    encoding: str = "utf-8",
    errors: str = "strict",
    line_length: float | int = LINE_LENGTH,
) -> bytes:
    """Serialize an XML tree.

    The iterator returned by this function yields the serialized XML
    piece by piece.

    Parameters
    ----------
    tree
        The XML tree to serialize.

    Returns
    -------
    Iterator[str]
        An iterator that yields the serialized XML piece by piece.
    """
    buffer = io.BytesIO()
    root: lxml.etree._Element
    preceding_siblings: cabc.Iterable[lxml.etree._Comment]
    following_siblings: cabc.Iterable[lxml.etree._Comment]
    if isinstance(tree, lxml.etree._ElementTree):
        root = tree.getroot()
        preceding_siblings = reversed(list(root.itersiblings(preceding=True)))
        following_siblings = root.itersiblings()
    else:
        root = tree
        preceding_siblings = ()
        following_siblings = ()

    pos = 0
    for i in preceding_siblings:
        pos = _serialize_comment(
            buffer, i, encoding=encoding, errors=errors, pos=pos, indent=0
        )

    _serialize_element(
        buffer,
        root,
        0,
        encoding=encoding,
        errors=errors,
        line_length=line_length,
        pos=pos,
    )
    if (root.tail or "").strip():
        pos = _serialize_text(
            buffer,
            root.tail,
            encoding=encoding,
            errors=errors,
            pos=pos,
            multiline=True,
        )

    for i in following_siblings:
        pos = _serialize_comment(
            buffer, i, encoding=encoding, errors=errors, pos=pos, indent=0
        )

    buffer.write(b"\n")
    return buffer.getvalue()


def _declare(encoding: str) -> bytes:
    return b"".join(
        (
            b'<?xml version="1.0" encoding="',
            encoding.upper().encode("ascii"),
            b'"?>',
            os.linesep.encode("ascii"),
        )
    )


def _escape(string: str, *, pattern: re.Pattern[str] = P_ESCAPE_TEXT) -> str:
    return pattern.sub(_escape_char, string)


def _escape_char(
    match: re.Match[str], *, ord_low: int = ord(" "), ord_high: int = ord("~")
) -> str:
    char = match.group(0)
    assert len(char) == 1
    if ord_low <= ord(char) <= ord_high:
        return "&{};".format(html.entities.codepoint2name[ord(char)])
    return f"&#x{ord(char):X};"


def _unmapped_attrs(
    nsmap: cabc.Mapping[str, str], element: lxml.etree._Element
) -> cabc.Iterator[tuple[str, str]]:
    if element.getparent() is None:
        parent_ns = set()
    else:
        parent_ns = set(element.getparent().nsmap)

    attribs = dict(element.items())
    try:
        yield (
            _unmap_namespace(nsmap, "xmi:version"),
            _escape(attribs.pop("{http://www.omg.org/XMI}version")),
        )
    except KeyError:
        pass

    for attr, value in element.nsmap.items():
        if attr in parent_ns:
            continue
        yield (f"xmlns:{attr}", value)

    for attr, value in attribs.items():
        yield (_unmap_namespace(nsmap, attr), _escape(value))


def _serialize_comment(
    buffer: _HasWrite,
    comment: lxml.etree._Comment,
    /,
    *,
    encoding: str,
    errors: str,
    pos: int,
    indent: int,
) -> int:
    assert isinstance(comment, lxml.etree._Comment)

    buffer.write(LINESEP)
    buffer.write(INDENT * indent)
    buffer.write(b"<!--")
    pos = _serialize_text(
        buffer,
        comment.text,
        encoding=encoding,
        errors=errors,
        pos=len(INDENT) * indent,
        pattern=re.compile(r">"),
    )
    buffer.write(b"-->")

    if (comment.tail or "").strip():
        pos = _serialize_text(
            buffer, comment.tail, encoding=encoding, errors=errors, pos=pos
        )
    else:
        buffer.write(LINESEP)
        buffer.write(INDENT * indent)
        pos = len(INDENT) * indent
    return pos


def _serialize_element(
    buffer: _HasWrite,
    element: lxml.etree._Element,
    indent: int,
    *,
    encoding: str,
    errors: str,
    pos: int = 0,
    line_length: float | int,
) -> int:
    assert isinstance(element, lxml.etree._Element)
    nsmap = dict((v, k) for k, v in element.nsmap.items())
    buffer.write(b"<")
    tag = _unmap_namespace(nsmap, element.tag).encode(encoding, errors)
    buffer.write(tag)

    pos += 1 + len(tag)
    attr_indent = INDENT * (indent + 2)
    force_break = False
    for attr, value in _unmapped_attrs(nsmap, element):
        if pos > line_length or force_break:
            buffer.write(LINESEP)
            buffer.write(attr_indent)
            pos = len(attr_indent)
            force_break = False
        else:
            buffer.write(b" ")
            pos += 1

        buffer.write(attr.encode(encoding, errors))
        buffer.write(b'="')
        buffer.write(value.encode(encoding, errors))
        buffer.write(b'"')
        pos += len(attr) + len(value) + 3

        if element.getparent() is None and attr == "id":
            force_break = True

    if (
        element.text is None
        and len(element) == 0
        and element.tag not in ALWAYS_EXPANDED_TAGS
    ):
        buffer.write(b"/>")
        return pos + 2
    buffer.write(b">")

    child_indent = INDENT * (indent + 1)
    if (element.text or "").strip():
        pos = _serialize_text(
            buffer,
            element.text,
            encoding=encoding,
            errors=errors,
            pos=pos,
            multiline=True,
        )
        text_content = True
    else:
        text_content = False

    for child in element:
        if not text_content:
            buffer.write(LINESEP)
            buffer.write(child_indent)
            pos = len(child_indent)

        pos = _serialize_element(
            buffer,
            child,
            indent + 1,
            encoding=encoding,
            errors=errors,
            pos=pos,
            line_length=line_length,
        )
        if (element.tail or "").strip():
            pos = _serialize_text(
                buffer, element.tail, encoding=encoding, errors=errors, pos=pos
            )
            text_content = True
        else:
            text_content = False

    if len(element) > 0 and not text_content:
        buffer.write(LINESEP)
        buffer.write(INDENT * indent)
        pos = len(INDENT) * indent

    buffer.write(b"</")
    buffer.write(tag)
    buffer.write(b">")
    return pos + len(tag) + 3


def _serialize_text(
    buffer: _HasWrite,
    text: str,
    /,
    *,
    encoding: str,
    errors: str,
    pos: int,
    multiline: bool = False,
    pattern: re.Pattern[str] = P_ESCAPE_TEXT,
) -> int:
    i, line = 0, ""
    for i, line in enumerate(text.split("\n")):
        if multiline and i:
            buffer.write(LINESEP)
        buffer.write(_escape(line, pattern=pattern).encode(encoding, errors))
    return len(line) + bool(i) * pos


def _unmap_namespace(nsmap: cabc.Mapping[str, str], name: str) -> str:
    match = P_NAME.search(name)
    assert match is not None
    if ns := match.group(1) or "":
        try:
            ns = nsmap[ns]
        except KeyError:
            raise ValueError(f"Namespace not found: {ns!r}") from None
        assert ns

    tag = match.group(2)
    assert tag

    return "".join((ns, ns and ":", tag))



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The AIRD parser and various other diagramming related tools.

This module is used to enumerate, access and export the diagrams from
Capella projects.  The JSON output it produces can be fed for example
into the :mod:`capellambse.svg` module for conversion to SVG.
"""
from .vector2d import *  # isort: skip
from .capstyle import *
from .diagram import *
from .json_enc import *
from .parser import *



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for the System Analysis layer.

This is normally the place to declare data used in the model for e.g.
functions, actors etc. which is best presented in a glossary document.

.. diagram:: [CDB] SA ORM
"""
import operator

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, capellacore, cs, fa, interaction
from . import oa

XT_ARCH = "org.polarsys.capella.core.data.ctx:SystemAnalysis"
XT_CAP_INV = "org.polarsys.capella.core.data.ctx:CapabilityInvolvement"


@c.xtype_handler(XT_ARCH)
class SystemFunction(c.GenericElement):
    """A system function."""

    _xmltag = "ownedFunctions"

    inputs = c.ProxyAccessor(fa.FunctionInputPort, aslist=c.ElementList)
    outputs = c.ProxyAccessor(fa.FunctionOutputPort, aslist=c.ElementList)
    is_leaf = property(lambda self: not self.functions)
    realized_operational_activities = c.ProxyAccessor(
        oa.OperationalActivity,
        fa.FunctionRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )

    owner: c.Accessor
    functions: c.Accessor
    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class SystemFunctionPkg(c.GenericElement):
    """A function package that can hold functions."""

    _xmltag = "ownedFunctionPkg"

    functions = c.ProxyAccessor(SystemFunction, aslist=c.ElementList)
    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class Mission(c.GenericElement):
    """A mission."""

    _xmltag = "ownedMissions"


@c.xtype_handler(XT_ARCH)
class MissionPkg(c.GenericElement):
    """A system mission package that can hold missions."""

    _xmltag = "ownedMissionPkg"

    missions = c.ProxyAccessor(Mission, aslist=c.ElementList)
    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class SystemComponent(cs.Component):
    """A system component."""

    _xmltag = "ownedSystemComponents"

    allocated_functions = c.ProxyAccessor(
        SystemFunction,
        fa.XT_FCALLOC,
        follow="targetElement",
        aslist=c.ElementList,
    )
    realized_operational_entities = c.ProxyAccessor(
        oa.Entity,
        cs.ComponentRealization,
        aslist=c.ElementList,
        follow="targetElement",
    )


@c.xtype_handler(XT_ARCH)
class SystemComponentPkg(c.GenericElement):
    """A system component package."""

    _xmltag = "ownedSystemComponentPkg"

    components = c.ProxyAccessor(SystemComponent, aslist=c.ElementList)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class Capability(c.GenericElement):
    """A capability."""

    _xmltag = "ownedCapabilities"

    owned_chains = c.ProxyAccessor(fa.FunctionalChain, aslist=c.ElementList)
    involved_functions = c.ProxyAccessor(
        SystemFunction,
        interaction.XT_CAP2ACT,
        aslist=c.ElementList,
        follow="involved",
    )
    involved_chains = c.ProxyAccessor(
        fa.FunctionalChain,
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        follow="involved",
    )
    involved_components = c.ProxyAccessor(
        SystemComponent,
        XT_CAP_INV,
        follow="involved",
        aslist=c.MixedElementList,
    )
    realized_capabilities = c.ProxyAccessor(
        oa.OperationalCapability,
        interaction.XT_CAP_REAL,
        follow="targetElement",
        aslist=c.ElementList,
    )

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.ProxyAccessor(interaction.Scenario, aslist=c.ElementList)
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityPkg(c.GenericElement):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.ProxyAccessor(Capability, aslist=c.ElementList)

    packages: c.Accessor


class SystemAnalysis(crosslayer.BaseArchitectureLayer):
    """Provides access to the SystemAnalysis layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        SystemComponent,
        attributes={"is_actor": False},
        rootelem=SystemComponentPkg,
    )
    root_function = c.ProxyAccessor(SystemFunction, rootelem=SystemFunctionPkg)

    function_package = c.ProxyAccessor(SystemFunctionPkg)
    capability_package = c.ProxyAccessor(CapabilityPkg)
    component_package = c.ProxyAccessor(SystemComponentPkg)
    mission_package = c.ProxyAccessor(MissionPkg)

    all_functions = c.ProxyAccessor(
        SystemFunction, deep=True, aslist=c.ElementList
    )
    all_capabilities = c.ProxyAccessor(
        Capability, deep=True, aslist=c.ElementList
    )
    all_components = c.ProxyAccessor(
        SystemComponent, deep=True, aslist=c.ElementList
    )
    all_actors = property(
        lambda self: self._model.search(SystemComponent).by_is_actor(True)
    )
    all_missions = c.ProxyAccessor(Mission, deep=True, aslist=c.ElementList)

    actor_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=SystemComponentPkg,
    )
    component_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=[SystemComponentPkg, SystemComponent],
        deep=True,
    )

    all_function_exchanges = c.ProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[SystemFunctionPkg, SystemFunction],
        deep=True,
    )
    all_component_exchanges = c.ProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        deep=True,
    )

    diagrams = diagram.DiagramAccessor(
        "System Analysis", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    Capability,
    "inheritance",
    c.ProxyAccessor(
        Capability,
        interaction.XT_CAP_GEN,
        follow="super",
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    SystemFunction,
    "owner",
    c.CustomAccessor(
        SystemComponent,
        operator.attrgetter("_model.sa.all_components"),
        matchtransform=operator.attrgetter("allocated_functions"),
    ),
)
c.set_accessor(
    SystemFunction,
    "packages",
    c.ProxyAccessor(
        SystemFunctionPkg,
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    oa.OperationalCapability,
    "realizing_capabilities",
    c.CustomAccessor(
        Capability,
        operator.attrgetter("_model.sa.all_capabilities"),
        matchtransform=operator.attrgetter("realized_capabilities"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    oa.Entity,
    "realizing_system_components",
    c.CustomAccessor(
        SystemComponent,
        operator.attrgetter("_model.sa.all_components"),
        matchtransform=operator.attrgetter("realized_operational_entities"),
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    oa.OperationalActivity,
    "realizing_system_functions",
    c.CustomAccessor(
        SystemFunction,
        operator.attrgetter("_model.sa.all_functions"),
        matchtransform=operator.attrgetter("realized_operational_activities"),
        aslist=c.ElementList,
    ),
)
c.set_self_references(
    (MissionPkg, "packages"),
    (SystemComponentPkg, "packages"),
    (SystemFunction, "functions"),
    (SystemFunctionPkg, "packages"),
)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from capellambse.loader import xmltools

from .. import common as c
from .. import modeltypes
from . import capellacommon, capellacore


def _allocated_exchange_items(
    obj: c.GenericElement,
) -> c.ElementList[c.GenericElement]:
    try:
        return obj.exchange_items
    except AttributeError:
        pass

    try:
        return obj.allocated_exchange_items
    except AttributeError:
        pass

    raise TypeError(
        f"Unhandled exchange type: {type(obj).__name__}"
    )  # pragma: no cover


def _search_all_exchanges(
    obj: c.GenericElement,
) -> c.ElementList[c.GenericElement]:
    from . import fa

    return obj._model.search(fa.ComponentExchange, fa.FunctionalExchange)


@c.xtype_handler(None)
class Class(c.GenericElement):
    """A Class."""

    _xmltag = "ownedClasses"

    inheritance = c.ProxyAccessor(capellacore.Generalization)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )


@c.xtype_handler(None)
class DataPkg(c.GenericElement):
    """A data package that can hold classes."""

    classes = c.ProxyAccessor(Class, aslist=c.ElementList)
    packages: c.Accessor


@c.xtype_handler(None)
class ExchangeItemElement(c.GenericElement):
    """An ExchangeItemElement (proxy link)."""

    _xmltag = "ownedElements"

    abstract_type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    owner: c.Accessor


@c.xtype_handler(None)
class ExchangeItem(c.GenericElement):
    """An item that can be exchanged on an Exchange."""

    _xmltag = "ownedExchangeItems"

    type = xmltools.EnumAttributeProperty(
        "_element",
        "exchangeMechanism",
        modeltypes.ExchangeItemType,
        default="UNSET",
    )
    elements = c.ProxyAccessor(
        ExchangeItemElement,
        aslist=c.ElementList,
        follow_abstract=False,
    )
    exchanges = c.CustomAccessor(
        c.GenericElement,
        _search_all_exchanges,
        matchtransform=_allocated_exchange_items,
        aslist=c.ElementList,
    )


c.set_accessor(
    capellacore.Generalization, "super", c.AttrProxyAccessor(Class, "super")
)
c.set_accessor(
    DataPkg, "packages", c.ProxyAccessor(DataPkg, aslist=c.ElementList)
)
c.set_accessor(ExchangeItemElement, "owner", c.ParentAccessor(ExchangeItem))



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides easy access to the Polarsys Capella PVMT extensions."""

from __future__ import annotations

from capellambse import NAMESPACES

from . import exceptions
from .core import AttributeProperty, XMLDictProxy
from .types import (
    AppliedPropertyValueGroup,
    EnumerationPropertyType,
    select_property_loader,
)

XPTH_FIND_BY_XTYPE = '*[@xsi:type="{}"]'
X_PVMT = "org.polarsys.capella.core.data.capellacore:PropertyValuePkg"


class Group(XMLDictProxy):
    """PVMT Group."""

    scope = AttributeProperty("xml_element", "description")

    def __init__(self, *args, parent=None, **kwargs):
        super().__init__(
            *args, **kwargs, childtag="ownedPropertyValues", keyattr="id"
        )
        self.parent = parent

    @property
    def properties(self):
        """Return a list of all properties defined in this group."""
        return list(self.values())

    def _extract_value(self, element):
        prop = select_property_loader(element)
        prop.parent = self
        return prop

    def _prepare_child(self, element, key):
        raise NotImplementedError(
            "Adding new properties to a group is not implemented yet"
        )

    def _insert_value(self, element, value):
        raise NotImplementedError(
            "Modifying property definitions is not implemented yet"
        )


class Domain(XMLDictProxy):
    """A PVMT Domain."""

    def __init__(self, element, *args, parent=None, **kwargs):
        super().__init__(
            element,
            *args,
            **kwargs,
            childtag="ownedPropertyValueGroups",
            keyattr="id",
        )
        self.enums = _DomainEnums(element)
        self.parent = parent

    @property
    def groups(self):
        """Return a list of all property value groups in this domain."""
        return list(self.values())

    def _extract_value(self, element):
        group = Group.from_xml_element(element)
        group.parent = self
        return group

    def _prepare_child(self, element, key):
        raise NotImplementedError(
            "Adding groups to a domain is not implemented yet"
        )

    def _insert_value(self, element, value):
        raise NotImplementedError(
            "Modifying domain groups is not implemented yet"
        )


class _DomainEnums(XMLDictProxy):
    _extract_value = EnumerationPropertyType.from_xml_element  # type: ignore[assignment]

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            childtag="ownedEnumerationPropertyTypes",
            keyattr="id",
        )

    def _prepare_child(self, element, key):
        raise NotImplementedError(
            "Adding new enums to a domain is not implemented yet"
        )

    def _insert_value(self, element, value):
        raise NotImplementedError(
            "Modifying enum definitions is not implemented yet"
        )


class PVMTExtension(XMLDictProxy):
    """Facilitates access to property values."""

    def __init__(self, element, model=None):
        super().__init__(
            element, childtag="ownedPropertyValuePkgs", keyattr="id"
        )
        self.model = model

    @property
    def domains(self):
        """Return a list of all property value domains in the model."""
        return list(self.values())

    def get_element_pv(
        self, element, groupname, create=True
    ) -> AppliedPropertyValueGroup:
        """Return the named PVMT group on ``element``.

        Parameters
        ----------
        element
            An LXML element with property value groups.
        groupname
            The fully qualified name of the property value group, in the
            format "domain.group".
        create
            True to create (apply) the group if necessary.
        """
        if "." not in groupname:
            raise ValueError(
                "Please specify the fully qualified property value group name"
            )

        for child in element.iterchildren("ownedPropertyValueGroups"):
            if child.attrib["name"] == groupname:
                break
        else:
            if not create:
                raise exceptions.GroupNotAppliedError(
                    "Property value group {} was not applied to this element".format(
                        groupname
                    )
                ) from None
            child = AppliedPropertyValueGroup.applyto(self, element, groupname)

        return AppliedPropertyValueGroup(self, child, model=self.model)

    def _extract_value(self, element):
        return Domain(element, parent=self)

    def _prepare_child(self, element, key):
        raise NotImplementedError(
            "Adding domains to the model is not implemented yet"
        )

    def _insert_value(self, element, value):
        raise NotImplementedError("Modifying domains is not implemented yet")


def load_pvmt_from_model(model):
    """Load the Property Value management extension for the given model.

    This function is the main entry point for the ``pvmt`` module.  It
    should be called after constructing a ``MelodyLoader`` instance on
    the model file.  It will return a ``PVMTExtension`` object, which
    can be used to easily access the property values of the model given
    during intialization.
    """
    pkgs = model.xpath(
        XPTH_FIND_BY_XTYPE.format(X_PVMT), namespaces=NAMESPACES
    )
    if not pkgs:
        raise ValueError("Provided model does not have a PropertyValuePkg")
    return PVMTExtension(pkgs[0], model)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=abstract-method, useless-suppression
# For some reason, pylint in Github CI didn't get the memo that these aren't
# actually abstract methods. Other pylint installations seem to agree that
# implementing these methods isn't necessary. So we just ignore the warning
# about that here.
# TODO Revisit this decision some time in the future

from __future__ import annotations

import collections.abc as cabc
import itertools
import os
import pathlib
import typing as t
import urllib.parse

import requests

from capellambse import helpers, loader

from . import FileHandler


class DownloadStream(t.BinaryIO):
    __stream: cabc.Iterator[bytes]
    __buffer: memoryview

    def __init__(self, url: str, chunk_size: int = 1024 ** 2) -> None:
        self.url = url
        self.chunk_size = chunk_size

        response = requests.get(self.url, stream=True)
        if response.status_code == 404:
            raise FileNotFoundError(url)
        response.raise_for_status()
        self.__stream = response.iter_content(
            self.chunk_size, decode_unicode=False
        )
        self.__buffer = memoryview(b"")

    def __enter__(self) -> DownloadStream:
        return self

    def __exit__(self, *args: t.Any) -> None:
        self.close()

    def read(self, n: int = -1) -> bytes:
        if n == -1:
            return b"".join(itertools.chain((self.__buffer,), self.__stream))

        if not self.__buffer:
            try:
                self.__buffer = memoryview(next(self.__stream))
            except StopIteration:
                return b""

        chunk = bytes(self.__buffer[:n])
        self.__buffer = self.__buffer[n:]
        return chunk

    def readable(self) -> bool:  # pylint: disable=no-self-use
        return True

    def write(self, s: bytes | bytearray) -> int:
        raise TypeError("Cannot write to a read-only stream")

    def writable(self) -> bool:  # pylint: disable=no-self-use
        return False

    def close(self) -> None:
        del self.__stream
        del self.__buffer


class HTTPFileHandler(FileHandler):
    """A remote file handler that fetches files using HTTP GET."""

    def __init__(
        self,
        path: str | os.PathLike,
        username: str | None = None,
        password: str | None = None,
    ) -> None:
        if not isinstance(path, str):
            raise TypeError(
                f"HTTPFileHandler requires a str path, not {type(path).__name__}"
            )
        if bool(username) != bool(password):
            raise ValueError(
                "Either both username and password must be given, or neither"
            )
        super().__init__(path)

        self.session = requests.Session()
        if username and password:
            self.session.auth = (username, password)

    def get_model_info(self) -> loader.ModelInfo:
        assert isinstance(self.path, str)
        parts = urllib.parse.urlparse(self.path)
        return loader.ModelInfo(
            title=parts.path.rsplit("/", maxsplit=1)[-1],
            url=self.path,
        )

    def open(
        self,
        filename: pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        if "w" in mode:
            raise NotImplementedError("Cannot upload to HTTP(S) locations")
        assert isinstance(self.path, str)
        fname_str = str(helpers.normalize_pure_path(filename))
        return DownloadStream(  # type: ignore[abstract] # false-positive
            "/".join((self.path, fname_str))
        )

    def write_transaction(self, **kw: t.Any) -> t.NoReturn:
        raise NotImplementedError(
            "Write transactions for HTTP(S) are not implemented"
        )
